"""
Created on Mon April 06 2020

This python file containts the classes and methods for stochastic analysis of
any power system using one of the following methods:

Monte Carlo simulation

@author: Dr. Jose Nicolas Melchor Gutierrez
"""

import math
import random
import time
import numpy as np
from pyene.engines.pyene import pyeneClass as pe
from .pyene_Models import Networkmodel as NMod # Energy model in glpk

# TODO: inherit pyene parameters
class MonteCarloSimulation():
    ''' This class contains all methods to perform the Monte Carlo simulation 
    of any power system '''

    def __init__(self, **kwargs):
        ''' This class method initialise '''
        if 'totaltimesteps' in kwargs:
            self.totaltimesteps = kwargs.pop('totaltimesteps')
        else:
            self.totaltimesteps = 0
        
        if 'iterationsMC' in kwargs:
            self.iterationsMC = kwargs.pop('iterationsMC')
        else:
            self.iterationsMC = 0

    def calculaterandomfailureandrecover(self, ob=None):
        ''' This class method calculates the time to failure and time to 
        recover of the different elements in the system '''
        assert isinstance(ob, pe), "The object that has been passed is not an \
            instance of pyene"
        TTF_gen = [None] * ob.NM.Gen.get_NoCon() # Time to failure generators
        TTR_gen = [None] * ob.NM.Gen.get_NoCon() # Time to recover generators
        TTF_lines = [None] * ob.NM.ENetwork.get_NoBra() # Time to failure lines
        TTR_lines = [None] * ob.NM.ENetwork.get_NoBra() # Time to recover lines
        
        for xg in range(ob.NM.Gen.get_NoCon()):
            TTF_gen[xg] = math.floor(-ob.NM.Gen.Conv[xg].data['MTTF'] * \
                math.log(random.random()))
            TTR_gen[xg] = math.floor(-ob.NM.Gen.Conv[xg].data['MTTR'] * \
                math.log(random.random()))
        
        for xl in range(ob.NM.ENetwork.get_NoBra()):
            TTF_lines[xl] = math.floor(-ob.NM.ENetwork.Branch[xl].data['MTTF'] \
                * math.log(random.random()))
            TTR_lines[xl] = math.floor(-ob.NM.ENetwork.Branch[xl].data['MTTR'] \
                * math.log(random.random()))
        
        return TTF_gen, TTR_gen, TTF_lines, TTR_lines

    def calculatematrixfailureandrecover(self, ob=None, **kwargs):
        ''' This class method creates the on/off matrix of different elements 
        per timestep for the Monte Carlo simulation '''
        Status_Generators = np.zeros( (ob.NM.Gen.get_NoCon(), \
            self.totaltimesteps) )
        Status_Lines = np.zeros( (ob.NM.ENetwork.get_NoBra(), \
            self.totaltimesteps) )

        TTF_gen = kwargs.pop('TTF_gen')
        TTR_gen = kwargs.pop('TTR_gen')
        TTF_lines = kwargs.pop('TTF_lines')
        TTR_lines = kwargs.pop('TTR_lines')

        for xl in range(ob.NM.ENetwork.get_NoBra()):
            count_l = 0
            while count_l < self.totaltimesteps:
                if TTF_lines[xl] == 0:
                    Status_Lines[xl,:] = 1
                    count_l = self.totaltimesteps
                else:
                    for _ in range(TTF_lines[xl]):
                        if (count_l<self.totaltimesteps):
                            Status_Lines[xl,count_l] = 1
                            count_l = count_l + 1

                if TTR_lines[xl] == 0:
                    Status_Lines[xl,:] = 1
                    count_l = self.totaltimesteps
                else:
                    for _ in range(TTR_lines[xl]):
                        if (count_l<self.totaltimesteps):
                            Status_Lines[xl,count_l] = 0
                            count_l = count_l + 1
        
        for xg in range(ob.NM.Gen.get_NoCon()):
            count_g = 0
            while count_g < self.totaltimesteps:
                if TTF_gen[xg] == 0:
                    Status_Generators[xg,:] = 1
                    count_g = self.totaltimesteps
                else:
                    for _ in range(TTF_gen[xg]):
                        if (count_g<self.totaltimesteps):
                            Status_Generators[xg,count_g] = 1
                            count_g = count_g + 1

                if TTR_gen[xg] == 0:
                    Status_Generators[xg,:] = 1
                    count_g = self.totaltimesteps
                else:
                    for _ in range(TTR_gen[xg]):
                        if (count_g<self.totaltimesteps):
                            Status_Generators[xg,count_g] = 0
                            count_g = count_g + 1
        
        sum_lines = Status_Lines.sum(axis=0)
        sum_generators = Status_Generators.sum(axis=0)
        Intact_Hour = np.zeros( self.totaltimesteps )
        for xt in range(self.totaltimesteps):
            if sum_lines[xt] == ob.NM.ENetwork.get_NoBra() and \
                sum_generators[xt] == ob.NM.Gen.get_NoCon():
                Intact_Hour[xt] = 1
                
        return Status_Lines, Status_Generators, Intact_Hour

    def runmontecarlo(self, ob=None, **kwargs):
        ''' This class method runs the monte carlo simulation '''
        assert isinstance(ob, pe), "The object that has been passed is not an \
            instance of pyene"

        demand_profiles = kwargs.pop('dem_profiles')
        # Running the power system without failure
        OF_timesteps_intact = np.zeros(self.totaltimesteps)
        LC_timesteps_intact = np.zeros(self.totaltimesteps)
        ConvC_timesteps_intact = np.zeros(self.totaltimesteps)
        Model = NMod(ob.NM)

        start = time.time()

        for xstep in range(self.totaltimesteps):
            for xn in range(Model.NumberNodesPS):
                if Model.NumberDemScenarios == 0:
                    Model.MultScenariosDemand[0,xn] = demand_profiles[xstep]
                else:
                    Model.MultScenariosDemand[0, 0, xn] = demand_profiles[xstep]
            Model.optimisationNM()
            if not Model.FlagProblem and Model.FlagFeasibility:
                LoadCurtailment = Model.GetLoadCurtailmentSystemED()
            elif Model.FlagProblem and Model.FlagFeasibility:
                LoadCurtailment = Model.GetLoadCurtailmentNodes()
            ThermalGenerationCurtailment = \
                Model.GetThermalGenerationCurtailmentNodes()
            for xh in Model.LongTemporalConnections:
                for xn in range(Model.NumberNodesPS):
                    for xco in range(Model.NumberContingencies + 1):
                        for xt in range(Model.ShortTemporalConnections):
                            if Model.FlagFeasibility:
                                LC_timesteps_intact[xstep] = \
                                    LC_timesteps_intact[xstep] + \
                                    LoadCurtailment[xh, xt, xco, xn]
                for xn in range(Model.NumberConvGen):
                    for xco in range(Model.NumberContingencies + 1):
                        for xt in range(Model.ShortTemporalConnections):
                            if Model.FlagFeasibility and \
                                Model.NumberConvGen > 0:
                                ConvC_timesteps_intact[xstep] = \
                                    ConvC_timesteps_intact[xstep] + \
                                    ThermalGenerationCurtailment\
                                    [xh, xt, xco, xn]
            OF_timesteps_intact[xstep] = Model.GetObjectiveFunctionNM()
            # print(OF_timesteps_intact[xstep])
            # print(LC_timesteps_intact[xstep])
            # print(ConvC_timesteps_intact[xstep])
            # print(xstep)
            # print()

        # Running the power system with failures
        OF_timesteps_MC = np.zeros(self.totaltimesteps * self.iterationsMC)
        LC_timesteps_MC = np.zeros(self.totaltimesteps * self.iterationsMC)
        ConvC_timesteps_MC = np.zeros(self.totaltimesteps * self.iterationsMC)

        # start = time.time()
        for xit in range(self.iterationsMC):
            # Calculate failure and recover times
            failure_recover_elements = {
                'TTF_gen' : None,
                'TTR_gen' : None,
                'TTF_lines' : None,
                'TTR_lines' : None
            }
            failure_recover_elements['TTF_gen'], \
            failure_recover_elements['TTR_gen'], \
            failure_recover_elements['TTF_lines'], \
            failure_recover_elements['TTR_lines'] = \
                self.calculaterandomfailureandrecover(ob)
            
            status_elements = {
                'Status_Lines' : None,
                'Status_Generators' : None,
                'Intact_Hour' : None
            }
    
            # Determine the status of elements
            status_elements['Status_Lines'], \
            status_elements['Status_Generators'], \
            status_elements['Intact_Hour'] = \
                self.calculatematrixfailureandrecover(ob=ob, \
                **failure_recover_elements)
            for xstep in range(self.totaltimesteps):
                for xn in range(Model.NumberNodesPS):
                    if Model.NumberDemScenarios == 0:
                        Model.MultScenariosDemand[0,xn] = demand_profiles[xstep]
                    else:
                        Model.MultScenariosDemand[0, 0, xn] = demand_profiles[xstep]
                
                if status_elements['Intact_Hour'][xstep] == 0:
                    for xconv in range(Model.NumberConvGen):
                        if status_elements['Status_Generators'][xconv, xstep] \
                            == 0:
                            Model.ActiveConv[xconv] = False
                        else:
                            Model.ActiveConv[xconv] = True
                    for xl in range(Model.NumberLinesPS):
                        if status_elements['Status_Lines'][xl, xstep] == 0:
                            Model.ActiveBranches[0, xl] = False
                        else:
                            Model.ActiveBranches[0, xl] = True
                    Model.optimisationNM()
                    if not Model.FlagProblem and Model.FlagFeasibility:
                        LoadCurtailment = Model.GetLoadCurtailmentSystemED()
                    elif Model.FlagProblem and Model.FlagFeasibility:
                        LoadCurtailment = Model.GetLoadCurtailmentNodes()
                    ThermalGenerationCurtailment = \
                        Model.GetThermalGenerationCurtailmentNodes()
                    for xh in Model.LongTemporalConnections:
                        for xn in range(Model.NumberNodesPS):
                            for xco in range(Model.NumberContingencies + 1):
                                for xt in range(Model.ShortTemporalConnections):
                                    if Model.FlagFeasibility:
                                        LC_timesteps_MC[\
                                        xit*self.totaltimesteps + xstep] = \
                                            LC_timesteps_MC[\
                                            xit*self.totaltimesteps + xstep] + \
                                            LoadCurtailment[xh, xt, xco, xn]
                        for xn in range(Model.NumberConvGen):
                            for xco in range(Model.NumberContingencies + 1):
                                for xt in range(Model.ShortTemporalConnections):
                                    if Model.FlagFeasibility and \
                                        Model.NumberConvGen > 0:
                                        ConvC_timesteps_MC[\
                                        xit*self.totaltimesteps + xstep] = \
                                            ConvC_timesteps_MC[\
                                            xit*self.totaltimesteps + xstep] + \
                                            ThermalGenerationCurtailment\
                                            [xh, xt, xco, xn]
                    OF_timesteps_MC[xit*self.totaltimesteps + xstep] = \
                        Model.GetObjectiveFunctionNM()
                else:
                    LC_timesteps_MC[xit*self.totaltimesteps + xstep] =\
                        LC_timesteps_intact[xstep]
                    ConvC_timesteps_MC[xit*self.totaltimesteps + xstep] =\
                        ConvC_timesteps_intact[xstep]
                    OF_timesteps_MC[xit*self.totaltimesteps + xstep] =\
                        OF_timesteps_intact[xstep]
                    # print(OF_timesteps_MC[xit*self.totaltimesteps + xstep])
                    # print(LC_timesteps_MC[xit*self.totaltimesteps + xstep])
                    # print(ConvC_timesteps_MC[xit*self.totaltimesteps + \
                    #     xstep])
                    # print(xit*self.totaltimesteps + xstep)
                    # print()

        end = time.time()
        time1 = end - start
        print('Time Monte Carlo:')
        print('%.15f' %(time1))
        print('Mean Time Monte Carlo per iteration:')
        print('%.15f' %(time1/self.totaltimesteps))
        print('\n\n')
        
        np.set_printoptions(threshold=np.inf)
        print(LC_timesteps_MC)


        # from .pyeneO import PrintinScreen

        # PiS = PrintinScreen(ob)
        # PiS.PrintallResults(Model)





from libc.stdlib cimport malloc, free
include "glpk.pxi"
inf = float('inf')


GLP_BND_TYPES = {
        'free': GLP_FR,  # free (unbounded) variable
        'lower': GLP_LO,  # variable with lower bound
        'upper': GLP_UP,  # variable with upper bound
        'bounded': GLP_DB,  # double-bounded variable
        'fixed': GLP_FX,  # fixed variable
}

GLP_DIR = {
    'min': GLP_MIN, 'minimise': GLP_MIN,
    'max': GLP_MAX, 'maximise': GLP_MAX,
}


GLP_METH = {
    'primal': GLP_PRIMAL,  # use primal simplex
    'dualprimal': GLP_DUALP,   # use dual; if it fails, use primal
    'dual': GLP_DUAL,   # use dual simplex
}

GLP_STATUS = {
    'basic': GLP_BS, # basic variable
    'nonbasiclower': GLP_NL, # non-basic variable having active lower bound
    'nonbasicupper': GLP_NU, # non-basic variable having active upper bound
    'nonbasicfree': GLP_NF, # non-basic free variable
    'nonbasicfixed': GLP_NS, # non-basic fixed variable
}

cdef class GLPKSolver:
    cdef glp_prob* prob
    cdef glp_smcp smcp

    cdef readonly dict row_ids
    cdef readonly dict col_ids
    cdef readonly  dict stats

    def __cinit__(self):
        # create a new problem
        self.prob = glp_create_prob()
        self.row_ids = {}
        self.col_ids = {}

    def __init__(self, time_limit=None, iteration_limit=None, \
        message_level='error', simplex_method='dualprimal'):

        self.stats = None

        # Set solver options
        glp_init_smcp(&self.smcp)
        self.smcp.msg_lev = message_levels[message_level]
        if time_limit is not None:
            self.smcp.tm_lim = time_limit  # 5 second limit
        if iteration_limit is not None:
            self.smcp.it_lim = iteration_limit
        
        self.smcp.meth = GLP_METH[simplex_method]

        glp_term_hook(term_hook, NULL)

    def __dealloc__(self):
        # free the problem
        glp_delete_prob(self.prob)
        
    cpdef set_dir(self, direction):
        glp_set_obj_dir(self.prob, GLP_DIR[direction])

    cpdef int add_rows(self, str name, int num):
        """Add new rows to the linear programme."""
        if name in self.row_ids:
            raise KeyError(f'Row name "{name}" already exists.')
        cdef int idx = glp_add_rows(self.prob, num)
        self.row_ids[name] = idx
        return idx

    cpdef int add_cols(self, str name, int num):
        """Add new columns to the linear programme."""
        if name in self.col_ids:
            raise KeyError(f'Column name "{name}" already exists.')
        cdef int idx = glp_add_cols(self.prob, num)
        self.col_ids[name] = idx

        for i in range(num):
            self.set_col_bnds(name, i, 'lower', 0.0, DBL_MAX)
        return idx

    cpdef set_mat_row(self, str name, int row_offset, cols, values):
        cdef int* ind
        cdef double* val

        if len(cols) != len(values):
            raise ValueError(
                f'The length of the `cols` ({len(cols)}) and `values` ({len(values)}) arguments mut be the same.')

        # Calculate the actual row in the LP to update
        cdef int row = self.row_ids[name] + row_offset

        ind = <int*>malloc((1 + len(cols)) * sizeof(int))
        val = <double*>malloc((1 + len(cols)) * sizeof(double))
        for n, c in enumerate(cols):
            ind[1+n] = 1+c
            # TODO check for finite values
            val[1+n] = values[n]
            print(row, c, n, values[n])
        glp_set_mat_row(self.prob, row, len(cols), ind, val)
        #glp_set_row_bnds(self.prob, row, GLP_FX, 0.0, 0.0)

        free(ind)
        free(val)

    cpdef load_matrix(self, int ne, iapy, japy, arpy):
        cdef int* ia
        cdef int* ja
        cdef double* ar

        ia = <int*> malloc((ne + 1) * sizeof(int))
        ja = <int*> malloc((ne + 1) * sizeof(int))
        ar = <double*> malloc((ne + 1) * sizeof(double))

        for aux1 in range(ne):
            ia[aux1 + 1] = iapy[aux1]
            ja[aux1 + 1] = japy[aux1]
            ar[aux1 + 1] = arpy[aux1]
        
        glp_load_matrix(self.prob, ne, ia, ja, ar)
        
        free(ia)
        free(ja)
        free(ar)

    cpdef set_row_bnds(self, str name, int row_offset, type, double lb, double ub):
        # TODO check for finite values here
        # Calculate the actual row in the LP to update
        cdef int row = self.row_ids[name] + row_offset
        glp_set_row_bnds(self.prob, row, GLP_BND_TYPES[type], lb, ub)

    cpdef set_col_bnds(self, str name, int col_offset, type, double lb, double ub):
        # TODO check for finite values here
        # Calculate the actual col in the LP to update
        cdef int col = self.col_ids[name] + col_offset
        glp_set_col_bnds(self.prob, col, GLP_BND_TYPES[type], lb, ub)

    cpdef set_obj_coef(self, str name, int col_offset, double coef):
        # TODO check for finite vaues here
        # Calculate the actual col in the LP to update
        cdef int col = self.col_ids[name] + col_offset
        glp_set_obj_coef(self.prob, col, coef)

    cpdef int simplex(self):
        return glp_simplex(self.prob, &self.smcp)

    cpdef int status(self):
        status = glp_get_status(self.prob)

    cpdef double get_col_prim(self, str name, int col_offset):
        # Calculate the actual col in the LP to update
        cdef int col = self.col_ids[name] + col_offset
        return glp_get_col_prim(self.prob, col)

    cpdef double get_obj_val(self):
        # retrieve objective value
        return glp_get_obj_val(self.prob)
    
    cpdef set_col_stat(self, str name, int col_offset, type):
        # Declare variables as basic or non-basic
        cdef int col = self.col_ids[name] + col_offset
        glp_set_col_stat(self.prob, col, GLP_STATUS[type])
    
    cpdef double get_row_dual(self, str name, int row_offset):
        # Calculate the dual value of a contraint
        cdef int row = self.row_ids[name] + row_offset
        return glp_get_row_dual(self.prob, row)





# Python Energy & Network Engine
"""
Created on Thu Mar 29 14:04:58 2018

The python energy and networks engine (pyene) provides a portfolio of tools for
assessing energy systems, as well as interfaces to exchange information with
the FutureDAMS framework and other engines and software.

@author: Dr Eduardo Alejandro Martínez Ceseña
https://www.researchgate.net/profile/Eduardo_Alejandro_Martinez_Cesena
"""
# convert in or long into float before performing divissions
from __future__ import division
from pyomo.core import ConcreteModel, Constraint, Objective, Suffix, Var, \
NonNegativeReals, minimize
from pyomo.opt import SolverFactory
import numpy as np
from pyensys.engines.pyeneN import ENetworkClass as dn  # Network cengine
from pyensys.engines.pyeneE import EnergyClass as de  # Energy engine
from pyensys.engines.pyeneH import HydrologyClass as hn  # Hydrology engine
from pyensys.engines.pyeneR import RESprofiles as rn  # RES engine#
from pyensys.engines.pyeneO import pyeneSave
from pyensys.engines.pyene_Models import EnergyandNetwork, Networkmodel
import json
import os


class pyeneConfig():
    ''' Overall default configuration '''
    def __init__(self):
        from .pyeneE import pyeneEConfig
        from .pyeneN import pyeneNConfig
        from .pyeneR import pyeneRConfig
        from .pyeneH import pyeneHConfig
        from .pyeneO import pyeneOConfig

        self.EN = ENEConfig()  # pyene
        self.EM = pyeneEConfig()  # pyeneE - Energy
        self.NM = pyeneNConfig()  # pyeneN - Networks
        self.RM = pyeneRConfig()  # pyeneR - Renewables
        self.HM = pyeneHConfig()  # pyeneH - Hidrology
        self.OM = pyeneOConfig()  # pyeneO - Outputs


class ENEConfig():
    ''' Default configuration for the integrated model '''
    def __init__(self):
        # Chose to load data from file
        self.Penalty = 1000000
        # Option to overwrite parameters
        self.Overwrite = {
                'DemandX': [],  # Position of demand profiles to overwrite
                'DemandL': [],  # Link position-->profile
                'Demand': [],  # New demand time-series
                'RESX': [],  # Position of RES profiles to overwrite
                'RESL': [],  # Link position-->profile
                'RES': [],  # New RES time-series
                'HydroX': [],  # Position of Hydropower values to overwrite
                'HydroL': [],  # Link position-->profile
                'Hydro': []  # New hydropower time-series
                }
        self.solverselection = {
            'pyomo': False,
            'glpk': True
        }


class pyeneClass():
    def __init__(self, obj=None):
        ''' Initialise network class '''
        # Get default values
        if obj is None:
            obj = ENEConfig()

        # Copy attributes
        for pars in obj.__dict__.keys():
            setattr(self, pars, getattr(obj, pars))

        # Additional parameters
        self.s = {}
        self.p = {}

    def _AddPyeneCons(self, m):
        ''' Model additional variables '''
        m = self.addVars(m)

        #                             Constraints                             #
        m = self.EM.addCon(m)
        m = self.NM.addCon(m)
        m = self.HM.addCon(m)
        m = self.addCon(m)

        # IEEE paper control variables.
        m.pEeee = self.EM.Print['IEEE']

        return m

    def _Calculate_OFaux(self):
        WghtAgg = 0+self.EM.p['WghtFull']
        OFaux = np.ones(len(self.NM.get_ConS()), dtype=float)
        xp = 0
        for xn in range(self.EM.LL['NosBal']+1):
            aux = self.EM.tree['After'][xn][0]
            if aux != 0:
                for xb in range(aux, self.EM.tree['After'][xn][1]+1):
                    WghtAgg[xb] *= WghtAgg[xn]
            else:
                OFaux[xp] = WghtAgg[xn]
                xp += 1

        return OFaux

    def _CheckInE(self, EM):
        ''' Print initial assumptions '''
        if EM.settings['File']:
            print('Loading resolution tree from file')
        else:
            print('Predefined resolution tree')
            print(EM.Input_Data)

    def _CheckInN(self, NM):
        ''' Print initial assumptions '''
        if NM.settings['Losses']:
            print('Losses are considered')
        else:
            print('Losses are neglected')
        if NM.settings['Feasibility']:
            print('Feasibility constrants are included')
        else:
            print('Feasibility constraints are neglected')
        print('Demand multiplyiers ', NM.settings['Demand'])
        print('Secuity: ', NM.settings['Security'])

    def addCon(self, m):
        ''' Adding pyomo constraints'''
        # Link water consumption throughout different models
        if self.NM.hydropower['Number'] > 0:
            if self.HM.settings['Flag']:
                # Collect inputs for pyeneH from pyeneE and pyeneN
                m.cAHMIn1 = Constraint(self.s['LL'], self.EM.s['Vec'],
                                       rule=self.cAHMIn1_rule)
                if self.EM.settings['Vectors'] < self.p['NoHMin']:
                    m.cAHMIn2 = Constraint(self.s['LL'],
                                           range(self.EM.settings['Vectors'],
                                           self.p['NoHMin']), self.NM.s['Tim'],
                                           rule=self.cAHMIn2_rule)
                # Connect pyeneH and pyeneN
                if self.p['NoHydDown'] > 0:
                    m.cAHMOut1 = \
                        Constraint(self.s['LL'], range(self.p['NoHydDown']),
                                   self.NM.s['Tim'], rule=self.cAHMOut1_rule)

                m.cAHMOut2 = Constraint(self.s['LL'], range(self.p['NoHMout']),
                                        self.NM.s['Tim'],
                                        rule=self.cAHMOut2_rule)

            else:
                # Link pyeneE and pyeneN
                m.cAEMNM = Constraint(self.s['LL'], self.EM.s['Vec'],
                                      rule=self.cAEMNM_rule)

        m.dual = Suffix(direction=Suffix.IMPORT)

        return m

    def addPar(self, m):
        ''' Add pyomo parameters'''
        if self.HM.settings['Flag']:
            # The head may change per iteration
            # Hydro
            self.p['EffHydro'] = np.zeros(self.NM.hydropower['Number'],
                                          dtype=float)
            for xn in range(self.NM.hydropower['Number']):
                self.p['EffHydro'][xn] = self.HM.hydropower['Head'][xn] * \
                    self.HM.hydropower['Efficiency'][xn]*9.81/1000
            # Pumps
            self.p['EffPump'] = np.zeros(self.NM.pumps['Number'], dtype=float)
            for xn in range(self.NM.pumps['Number']):
                self.p['EffPump'][xn] = self.HM.pumps['Head'][xn] * \
                    self.HM.pumps['Efficiency'][xn]*9.81/1000

        return m

    def addSets(self, m):
        '''Add pyomo sets'''

        return m

    def addVars(self, m):
        ''' Add pyomo variables '''
        # Converting some parameters to variables
        del m.vEOut
        m.vEOut = Var(self.EM.s['Nodz'], self.EM.s['Vec'],
                      domain=NonNegativeReals, initialize=0.0)
        return m

    def build_Mod(self, m):
        ''' Build pyomo model '''

        # Sets
        m = self.EM.addSets(m)
        m = self.NM.addSets(m)
        m = self.HM.addSets(m)
        m = self.addSets(m)

        # Model Variables
        m = self.EM.addVars(m)
        m = self.NM.addVars(m)
        m = self.HM.addVars(m)

        # Parameters
        m = self.EM.addPar(m)
        m = self.NM.addPar(m)
        m = self.HM.addPar(m)
        m = self.addPar(m)

        return m

    def cAEMNM_rule(self, m, xL, xv):
        ''' Connecting  pyeneE and pyeneN (MW --> MW)'''
        # TODO: Send first half of the constraint to pyeneH
        return m.vEOut[self.p['pyeneE'][xL], xv] == \
            self.NM.In_From_EM(m, xL, xv)

    def cAHMIn1_rule(self, m, xL, xv):
        ''' Flows from pyeneE and pyeneHin (MW --> m^3/s)'''
        xp = self.p['pyeneHin'][xv][1]+xL*(self.NM.pumps['Number']+1)
        return m.vEOut[self.p['pyeneE'][xL], xv] == \
            sum((m.vHin[self.p['NoHMin']*xL+xv, xt] -
                 sum(m.vNPump[xp, xt] *
                     self.NM.ENetwork.get_Base() /
                     self.p['EffPump'][self.p['pyeneHin'][xv][1]]
                     for x in range(self.p['pyeneHin'][xv][0]))) *
                self.NM.scenarios['Weights'][xt]
                for xt in self.NM.s['Tim'])*self.p['EffHydro'][xv]

    def cAHMIn2_rule(self, m, xL, xv, xt):
        ''' Connecting  pyeneE and pyeneHin (MW --> m^3/s)'''
        return m.vHin[self.p['NoHMin']*xL+xv, xt] == \
            sum(m.vNPump[self.p['pyeneHin'][xv][1]+xL *
                         (self.NM.pumps['Number']+1)+1, xt] *
                self.NM.ENetwork.get_Base() /
                self.p['EffPump'][self.p['pyeneHin'][xv][1]]
                for x in range(self.p['pyeneHin'][xv][0]))

    def cAHMOut1_rule(self, m, xh, xv, xt):
        ''' pyeneH outputs taken from the aggregated downstream flows '''
        # Position of the hydropower plant
        xb = self.p['LLHydDown'][xv]
        # Node to be addressed
        xn = self.HM.hydropower['Node'][xb]-1

        return self.NM.In_From_HM(m, xh, xt, xb, self.p['EffHydro'][xb]) <= \
            sum(m.vHup[self.HM.p['ConRiver'][xh] +
                       self.HM.p['LLN2B1'][self.HM.p['LLN2B2'][xn, 3]+xd], xt]
                for xd in range(self.HM.p['LLN2B2'][xn, 2]))

    def cAHMOut2_rule(self, m, xh, xn, xt):
        ''' pyeneH outputs taken from nodal outputs '''
        # Position of hydro in pyeneH
        xb = self.p['LLHydOut'][xn][1]
        # Position of the pump
        xp = self.p['LLHPumpOut'][xn][1]

        return m.vHout[xn+xh*self.HM.nodes['OutNumber'], xt] >= \
            sum(self.NM.In_From_HM(m, xh, xt, xb, self.p['EffHydro'][xb])
                for x in range(self.p['LLHydOut'][xn][0])) + \
            sum(m.vNPump[1+xp+xh*(1+self.NM.pumps['Number']), xt] *
                self.NM.ENetwork.get_Base()/self.p['EffPump'][xp]
                for x in range(self.p['LLHPumpOut'][xn][0]))

    def CheckProfile(self, value):
        ''' Verify that the size of the profile makes sense'''
        if len(value) != self.NM.settings['NoTime']:
            if self.NM.settings['NoTime'] == 1:
                value = [np.mean(value)]
            elif self.NM.settings['NoTime'] == 2:
                value = [(sum(value[0:15])+sum(value[20:24]))/19,
                         np.mean(value[15:20])]
            else:
                raise RuntimeError('Incompatible demand profiles')

        return value

    def close_output_files(self):
        self.save_solution.close_output_files()

    def ESim(self, conf):
        ''' Energy only optimisation '''
        # Get energy object
        EM = de(conf.EM)

        # Initialise
        EM.initialise()

        # Build LP model
        EModel = self.SingleLP(EM)

        #                          Objective function                         #
        EModel.OF = Objective(rule=EM.OF_rule, sense=minimize)

        # Optimise
        opt = SolverFactory('glpk')

        # Print
        results = opt.solve(EModel)

        return (EM, EModel, results)

    def get_AllDemand(self, m, *varg, **kwarg):
        '''Get the demand'''
        # Specify buses
        if 'buses' in kwarg:
            auxbuses = kwarg.pop('buses')
        else:
            auxbuses = range(self.NM.ENetwork.get_NoBus())

        value = 0
        values = [0, 0]
        for xn in auxbuses:
            aux = self.get_Demand(m, xn+1, *varg, **kwarg)
            values[self.NM.ENetwork.Bus[xn].get_LT()] += aux
            value += aux

        return value, values

    def get_AllDemandCurtailment(self, m, *varg, **kwarg):
        '''Get the kWh that had to be curtailed from all buses'''
        # Specify buses
        if 'buses' in kwarg:
            auxbuses = kwarg.pop('buses')
        else:
            auxbuses = range(self.NM.ENetwork.get_NoBus())

        values = [0, 0]
        value = 0
        if self.NM.settings['Feasibility']:
            for xn in auxbuses:
                aux = self.get_DemandCurtailment(m, xn, *varg, **kwarg)
                values[self.NM.ENetwork.Bus[xn].get_LT()] += aux
            value = values[0]+values[1]

        return value, values

    def get_AllGeneration(self, m, *varg, **kwarg):
        ''' Get kWh for all generators for the whole period '''
        if 'All' in varg:
            aux = self.NM.Gen.get_GenAll()
        elif 'Conv' in varg:
            aux = self.NM.Gen.get_GenAllC()
        elif 'RES' in varg:
            aux = self.NM.Gen.get_GenAllR()
        elif 'Hydro' in varg:
            aux = self.NM.Gen.get_GenAllH()
        else:
            aux = self.NM.Gen.get_GenAll()

        value = 0
        for xn in aux:
            value += self.get_Generation(m, xn, *varg, **kwarg)

        return value

    def get_AllHydro(self, m):
        ''' Get surplus kWh from all hydropower plants '''
        value = 0
        for xi in range(self.EM.settings['Vectors']):
            value += m.vEOut[1, xi].value

        return value

    def get_AllLoss(self, m, *varg, **kwarg):
        '''Get the total losses'''
        value = 0
        if self.NM.settings['Losses']:
            for xb in self.NM.s['Bra']:
                value += self.get_Loss(m, xb, *varg, **kwarg)

        return value

    def get_AllPumps(self, m, *varg, **kwarg):
        ''' Get kWh consumed by all pumps '''
        value = 0
        for xp in range(self.NM.pumps['Number']):
            value += self.get_Pump(m, xp+1, *varg, **kwarg)

        return value

    def get_AllRES(self, m, *varg, **kwarg):
        ''' Total RES spilled for the whole period '''
        value = 0
        for xr in range(self.NM.RES['Number']):
            value += self.get_RES(m, xr+1, *varg, **kwarg)

        return value

    def get_Demand(self, m, bus, *varg, **kwarg):
        '''Get the kWh that had to be curtailed from a given bus'''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        value = 0
        xb = bus-1
        for xh in auxscens:
            acu = 0
            for xt in auxtime:
                acu += (self.NM.busData[xb]*self.NM.scenarios['Demand']
                        [xt+self.NM.busScenario[xb][xh]])*auxweight[xt]
            value += acu*auxOF[xh]
        value *= self.NM.ENetwork.get_Base()

        return value

    def get_DemandCurtailment(self, m, bus, *varg, **kwarg):
        '''Get the kWh that had to be curtailed from a given bus'''
        if self.NM.p['LLFea1'][bus] == 0:
            return 0

        value = 0
        if isinstance(m, ConcreteModel):
            (auxtime, auxweight, auxscens,
             auxOF) = self.get_timeAndScenario(*varg, **kwarg)

            if self.NM.settings['Feasibility']:
                for xh in auxscens:
                    acu = 0
                    for xt in auxtime:
                        acu += auxweight[xt] * \
                            m.vNFea[self.NM.get_ConFea(xh) + \
                                    self.NM.p['LLFea2'][bus], xt].value
                    value += acu*auxOF[xh]
                value *= self.NM.ENetwork.get_Base()
        elif isinstance(m, EnergyandNetwork) or isinstance(m, Networkmodel):
            if m.FlagProblem and m.FlagFeasibility:
                LoadCurtailment = m.GetLoadCurtailmentNodes()
            else:
                LoadCurtailment = m.GetLoadCurtailmentSystemED()
            for xs in m.LongTemporalConnections:
                for xt in range(m.ShortTemporalConnections):
                    for k in range(m.NumberContingencies + 1):
                        value += LoadCurtailment[xs, xt, k, bus]

        return value

    def get_Generation(self, m, index, *varg, **kwarg):
        ''' Get kWh for a single generator '''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        value = 0
        for xh in auxscens:
            acu = 0
            for xt in auxtime:
                acu += m.vNGen[self.NM.get_ConG(xh)+index, xt].value * \
                    auxweight[xt]
            value += acu*auxOF[xh]
        value *= self.NM.ENetwork.get_Base()

        return value

    def get_Hydro(self, m, index):
        ''' Get surplus kWh from specific site '''
        HydroValue = m.vEOut[1, index-1].value

        return HydroValue

    def get_HydroFlag(self, m, index):
        ''' Get surplus kWh from specific site '''
        cobject = getattr(m, 'cSoCBalance')
        aux = m.dual.get(cobject[1, index-1])
        if aux is None:
            HydroValue = False
        else:
            aux2 = -1*int(m.dual.get(cobject[1, index-1]))
            aux3 = self.Penalty/self.NM.ENetwork.get_Base()
            if aux2 > aux3:
                HydroValue = True
            else:
                HydroValue = False

        return HydroValue

    def get_HydroMarginal(self, m, index):
        ''' Get marginal costs for specific hydropower plant '''
        cobject = getattr(m, 'cSoCBalance')
        aux = m.dual.get(cobject[1, index-1])
        if aux is None:
            HydroValue = 0
        else:
            HydroValue = -1*int(m.dual.get(cobject[1, index-1]))

        return HydroValue

    def get_Loss(self, m, xb, *varg, **kwarg):
        '''Get losses for a given branch'''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        value = 0
        if self.NM.settings['Flag']:
            # If the electricity network has been modelled
            if self.NM.settings['Losses']:
                value = 0
                for xh in auxscens:
                    acu = 0
                    for xt in auxtime:
                        acu += m.vNLoss[self.NM.get_ConL(xh)+xb, xt].value * \
                            auxweight[xt]
                    value += acu*auxOF[xh]
                value *= self.NM.ENetwork.get_Base()
            elif self.NM.settings['Loss'] is not None:
                # TODO: Enable option
                import sys
                sys.exit('Options with network and predefined losses not yet enabled')

        elif self.NM.settings['Loss'] is not None:
            # If losses have been estimated
            import sys
            sys.exit('Options without network and predefined losses not yet enabled')

        import sys
        sys.exit('Just stop')
        return value

    def get_MeanHydroMarginal(self, m):
        value = 0
        for xi in range(self.EM.settings['Vectors']):
            value += self.get_HydroMarginal(m, xi+1)
        return value/self.EM.settings['Vectors']

    def get_NetDemand(self, m, auxFlags, *varg, **kwarg):
        ''' Get MWh consumed by pumps, loads, losses '''
        value = 0
        if auxFlags[0]:  # Demand
            value += self.get_AllDemand(m, *varg, **kwarg)

        if auxFlags[1]:  # Pumps
            value += self.get_AllPumps(m, *varg, **kwarg)

        if auxFlags[2]:  # Loss
            value += self.get_AllLoss(m, *varg, **kwarg)

        if auxFlags[3]:  # Curtailment
            value += self.get_AllDemandCurtailment(m, *varg, **kwarg)[0]

        if auxFlags[4]:  # Spill
            value += self.get_AllRES(m, *varg, **kwarg)

        return value

    def get_OFpart(self, m, xg, *varg, **kwarg):
        ''' Get components of the objective function '''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        value = sum(sum(m.vNGCost[self.NM.get_ConC(xh)+xg, xt].value
                        for xt in auxtime)*auxOF[xh] for xh in auxscens)

        return value

    def get_OFparts(self, m, auxFlags, *varg, **kwarg):
        # TODO: Validate
        ''' Get components of the objective function '''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        value = 0
        if auxFlags[0]:  # Conventional generation
            for x in self.NM.Gen.Conv:
                value += sum(sum(m.vNGCost[self.NM.get_ConC(xh)+x.get_vNGen(),
                                           xt].value for xt in
                                 auxtime)*auxOF[xh] for xh in auxscens)
        if auxFlags[1]:  # RES generation
            for x in self.NM.Gen.RES:
                value += sum(sum(m.vNGCost[self.NM.get_ConC(xh)+x.get_vNGen(),
                                           xt].value for xt in
                                 auxtime)*auxOF[xh] for xh in auxscens)

        if auxFlags[2]:  # Hydro generation
            for x in self.NM.Gen.Hydro:
                value += sum(sum(m.vNGCost[self.NM.get_ConC(xh)+x.get_vNGen(),
                                           xt].value for xt in
                                 auxtime)*auxOF[xh] for xh in auxscens)

        if auxFlags[3]:  # Pumps
            value -= sum(sum(self.NM.pumps['Value'][xdl] *
                             self.NM.ENetwork.get_Base() *
                             sum(m.vNPump[self.NM.get_ConP(xh)+xdl+1,
                                          xt].value
                                 for xt in auxtime)
                             for xdl in self.NM.s['Pump']) *
                         auxOF[xh] for xh in auxscens)

        if auxFlags[4]:  # Curtailment
            value += sum(sum(sum(m.vNFea[self.NM.get_ConFea(xh)+xf, xt].value
                                 for xf in self.NM.s['Fea'])*self.Penalty
                             for xt in auxtime) *
                         auxOF[xh] for xh in auxscens)

        return value

    def get_Pump(self, m, index, *varg, **kwarg):
        ''' Get kWh consumed by a specific pump '''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)
        value = 0
        for xh in auxscens:
            acu = 0
            for xt in auxtime:
                acu += auxweight[xt] * \
                    m.vNPump[self.NM.get_ConP(xh)+index, xt].value
            value += acu*auxOF[xh]
        value *= self.NM.ENetwork.get_Base()

        return value

    def get_RES(self, m, index, *varg, **kwarg):
        ''' Spilled kWh of RES for the whole period'''
        (auxtime, auxweight, auxscens,
         auxOF) = self.get_timeAndScenario(*varg, **kwarg)

        xg = index-1
        value = 0
        for xh in auxscens:
            acu = 0
            for xt in auxtime:
                acu += ((self.NM.RES['Max'][xg]*self.NM.scenarios['RES']
                         [self.NM.resScenario[xg][xh]+xt] -
                         m.vNGen[self.NM.connections['Generation'][xh] +
                                 self.NM.Gen.get_vNGenR(xg), xt].value) *
                        auxweight[xt])
            value += acu*auxOF[xh]
        value *= self.NM.ENetwork.get_Base()

        return value

    def get_timeAndScenario(self, *varg, **kwarg):
        # Specify times
        if 'times' in kwarg:
            auxtime = kwarg.pop('times')
        else:
            auxtime = self.NM.s['Tim']

        # Remove weights
        if 'snapshot' in varg:
            auxweight = np.ones(len(self.NM.s['Tim']), dtype=int)
            auxOF = np.ones(len(self.NM.get_ConS()), dtype=int)
        else:
            auxweight = self.NM.scenarios['Weights']
            auxOF = self.OFaux

        # Specify scenario
        if 'scens' in kwarg:
            auxscens = kwarg.pop('scens')
        else:
            auxscens = self.NM.get_ConS()

        return (auxtime, auxweight, auxscens, auxOF)

    def getClassInterfaces(self):
        ''' Return calss to interface with pypsa and pypower'''
        from .pyeneI import EInterfaceClass

        return EInterfaceClass()

    def getClassOutputs(self, obj=None):
        ''' Return calss to produce outputs in H5 files'''
        from .pyeneO import pyeneHDF5Settings

        return pyeneHDF5Settings(obj)

    def getClassRenewables(self, obj=None):
        ''' Return calss to produce RES time series'''
        from .pyeneR import RESprofiles

        return RESprofiles(obj)

    def HSim(self, conf):
        ''' Hydrology only optimisation '''
        # Get network object
        HM = hn(conf.HM)

        # Initialise
        HM.initialise()

        # Build LP model
        HModel = self.SingleLP(HM)

        #                          Objective function                         #
        HModel.OF = Objective(rule=HM.OF_rule, sense=minimize)

        # Optimise
        opt = SolverFactory('glpk')

        # Print results
        results = opt.solve(HModel)

        return (HM, HModel, results)

    def initialise(self, conf=None, path=None):
        ''' Initialise energy and networks simulator '''
        if path is None:
            # Creat objects
            self.EM = de(conf.EM)

            self.NM = dn(conf.NM)

            # Adding hydro to the energy balance tree
            self.EM.settings['Fix'] = True,  # Force a specific number of vectors
            # Get number of vectors
            if conf.HM.settings['Flag']:
                # one per input node of the water network
                self.EM.settings['Vectors'] = len(conf.HM.nodes['In'])
            else:
                # One per hydropower plant
                self.EM.settings['Vectors'] = self.NM.hydropower['Number']
            self.s['Hydro'] = range(self.NM.hydropower['Number'])

            # Initialise energy balance model
            self.EM.initialise()

            # Get number of required network model instances
            NoNM = (1+self.EM.tree['Time'][self.EM.size['Periods']][1] -
                    self.EM.tree['Time'][self.EM.size['Periods']][0])

            # Add time steps
            aux = self.EM.size['Scenarios']
            self.NM.scenarios['Number'] = aux
            self.NM.scenarios['NoDem'] = aux


            if self.NM.scenarios['NoDem'] > 0:
                aux = self.NM.scenarios['Demand']
                self.NM.scenarios['Demand'] = \
                    np.ones(self.NM.settings['NoTime'] *
                            self.NM.scenarios['NoDem'], dtype=float)
                for x in range(len(aux)):
                    self.NM.scenarios['Demand'][x] = aux[x]

            # Initialise RES
            if self.NM.RES['Number'] > 0:
                self.NM.scenarios['RES'] = \
                    np.zeros(self.NM.settings['NoTime']*self.NM.scenarios['NoRES'],
                             dtype=float)

            # Initialise network model
            self.NM.initialise(rn(conf.RM))

            # Add connections between energy balance and network models
            self.NM.set_ConS(range(NoNM))
            auxFlow = np.zeros(NoNM, dtype=int)
            auxVol = np.zeros(NoNM, dtype=int)
            auxLoss = np.zeros(NoNM, dtype=int)
            auxGen = np.zeros(NoNM, dtype=int)
            auxCost = np.zeros(NoNM, dtype=int)
            auxPump = np.zeros(NoNM, dtype=int)
            auxFea = np.zeros(NoNM, dtype=int)
            # Location of each instance
            aux = self.NM.get_ConB()
            for xc in self.NM.get_ConS():
                auxFlow[xc] = xc*(self.NM.NoBranch)
                auxVol[xc] = xc*(self.NM.NoBuses)
                auxLoss[xc] = xc*(aux)
                auxGen[xc] = xc*(self.NM.Gen.get_NoGen())
                auxCost[xc] = xc*self.NM.Gen.get_NoGen()
                auxPump[xc] = xc*(self.NM.pumps['Number']+1)
                auxFea[xc] = xc*self.NM.NoFea

            self.NM.set_ConF(auxFlow)
            self.NM.set_ConV(auxVol)
            self.NM.set_ConL(auxLoss)
            self.NM.set_ConG(auxGen)
            self.NM.set_ConC(auxCost)
            self.NM.set_ConP(auxPump)
            self.NM.set_ConFea(auxFea)
            del auxFlow, auxVol, auxLoss, auxGen, auxPump, auxFea

            # Build LL to link the models through hydro consumption
            # Adding connections to pyeneE
            self.p['Number'] = NoNM
            self.s['LL'] = range(self.p['Number'])
            self.p['pyeneE'] = np.zeros(self.p['Number'], dtype=int)
            for xc in self.s['LL']:
                self.p['pyeneE'][xc] = \
                    self.EM.tree['Time'][self.EM.size['Periods']][0]+xc

            # Adding connections to pyeneN
            self.p['pyeneN'] = np.zeros(self.p['Number'], dtype=int)
            aux = self.NM.Gen.get_NoGen()+1
            for xc in self.s['LL']:
                self.p['pyeneN'][xc] = self.NM.get_ConG(xc)+aux

            # Create hydraulic data
            if conf.HM.settings['Flag']:
                # set periods
                conf.HM.settings['NoTime'] = self.NM.settings['NoTime']
                # adjust length of time period
                conf.HM.settings['seconds'] *= self.NM.scenarios['Weights'][0]
                # Remove fixed water flows
                conf.HM.settings['In'] = []
                # Define scenarios
                conf.HM.connections['Number'] = NoNM
                # Connect beginning and end of each scenario
                conf.HM.connections['LinksF'] = np.zeros((NoNM, 2), dtype=int)
                conf.HM.connections['LinksT'] = np.ones((NoNM, 2), dtype=int)
                for xh in range(NoNM):
                    conf.HM.connections['LinksF'][xh][0] = xh
                    conf.HM.connections['LinksT'][xh][0] = xh

            # Create hydraulic model
            self.HM = hn(conf.HM)

            # Initialise model
            self.HM.initialise()

            # TODO: Replace LL
            # Create LL
            if self.HM.settings['Flag']:
                # Inputs to pyeneH
                self.p['NoHMin'] = len(self.HM.nodes['In'])
                self.p['pyeneHin'] = np.zeros((self.p['NoHMin'], 2), dtype=int)
                # Adding pump inputs
                for x in range(self.NM.pumps['Number']):
                    xp = self.HM.pumps['To'][x]
                    if xp != 0:
                        xn = self.EM.settings['Vectors']-1

                        while self.HM.nodes['In'][xn] != xp:
                            xn -= 1
                        self.p['pyeneHin'][xn][0:2] = [1, x]

                # Outputs from pyeneH
                # Hydropower units with rivers connected downstream - The info
                # is calculated based on the summation of the flows
                self.p['NoHydDown'] = 0
                self.p['LLHydDown'] = np.zeros(self.NM.hydropower['Number'],
                                               dtype=int)
                for x in range(self.NM.hydropower['Number']):
                    xn = self.HM.hydropower['Node'][x]
                    if self.HM.p['LLN2B2'][xn-1, 2] > 0:
                        self.p['LLHydDown'][self.p['NoHydDown']] = x
                        self.p['NoHydDown'] += 1
                self.p['LLHydDown'] = self.p['LLHydDown'][0:self.p['NoHydDown']]

                # Hydropower units without downstream rivers or other devices that
                # require comparison with water outpouts
                self.p['NoHMout'] = len(self.HM.nodes['Out'])
                self.p['LLHydOut'] = np.zeros((self.p['NoHMout'], 2), dtype=int)
                self.s['LLHydOut'] = np.zeros(self.p['NoHMout'], dtype=int)
                xh = 0
                for x in range(self.NM.hydropower['Number']):
                    xn1 = self.HM.hydropower['Node'][x]-1
                    xn2 = self.HM.connections['NodeSeqOut'][xn1]
                    if xn2 != 0:
                        self.p['LLHydOut'][xn2-1][:] = [1, x]
                        self.s['LLHydOut'][xh] = x
                        xh += 1
                self.s['LLHydOut'] = self.s['LLHydOut'][0:xh]

                # Avoid double counting
                for x in self.p['LLHydDown']:
                    xn = self.HM.hydropower['Node'][x]-1
                    xn = self.HM.connections['NodeSeqOut'][xn]
                    self.p['LLHydOut'][xn-1][:] = [0, 0]
                # Pumps
                self.p['LLHPumpOut'] = np.zeros((self.p['NoHMout'], 2), dtype=int)
                for x in range(self.NM.pumps['Number']):
                    xn1 = self.HM.pumps['From'][x]-1
                    xn2 = self.HM.connections['NodeSeqOut'][xn1]
                    if xn2 != 0:
                        self.p['LLHPumpOut'][xn2-1][:] = [1, x]
        else:
            _, file_extension = os.path.splitext(path)
            
            from .pyene_Models import models
            self.mod_cython_cpp = models()
            if file_extension == ".xlsx":
                from .pyeneI import excel2pyene as e2p
                read = e2p()
                read.read_excel(energy_file=path, model=self.mod_cython_cpp)
            self.mod_cython_cpp.initialise()
            outpput_file_name = "outputs"
            for options in self.mod_cython_cpp.model_options:
                if options.name == "output file name":
                    outpput_file_name = options.value

            self.save_solution = pyeneSave()
            self.save_solution.initialise(model=self.mod_cython_cpp, dir=os.path.dirname(path), name_file=outpput_file_name)

    def NSim(self, conf):
        ''' Network only optimisation '''
        # Get network object
        NM = dn(conf.NM)

        # Initialise
        NM.initialise(rn(conf.RM))

        # Build LP model
        NModel = self.SingleLP(NM)

        #                          Objective function                         #
        NModel.OF = Objective(rule=NM.OF_rule, sense=minimize)

        # Optimise
        opt = SolverFactory('glpk')

        # Print results
        results = opt.solve(NModel)

        return (NM, NModel, results)

    def OF_rule(self, m):
        ''' Objective function for energy and networks model'''
        return sum((sum((sum(m.vNGCost[self.NM.get_ConC(xh)+xg, xt]
                             for xg in self.NM.s['Gen']) +
                         sum(m.vNFea[self.NM.get_ConFea(xh)+xf, xt]
                             for xf in self.NM.s['Fea']) *
                         self.Penalty)*self.NM.scenarios['Weights'][xt]
                        for xt in self.NM.s['Tim']) -
                    sum(self.NM.pumps['Value'][xdl] *
                        self.NM.ENetwork.get_Base() *
                        sum(m.vNPump[self.NM.get_ConP(xh)+xdl+1, xt] *
                            self.NM.scenarios['Weights'][xt]
                            for xt in self.NM.s['Tim'])
                        for xdl in self.NM.s['Pump'])) *
                   self.OFaux[xh] for xh in self.NM.get_ConS()) + \
            m.vHpenalty

    def Print_ENSim(self, m):
        ''' Print results '''
        
        from pyensys.engines.pyeneP import PrintClass
        prnt = PrintClass(m, self)
        
        self.EM.print(m)
        for xh in range(self.p['Number']):
            self.NM.print(m, [xh], prnt)
            self.HM.print(m, [xh])
            print()

        print('Water outputs:')
        for xn in self.EM.s['Nodz']:
            for xv in self.EM.s['Vec']:
                aux = m.vEOut[xn, xv].value
                print("%8.4f " % aux, end='')
            print('')
        print('Water inputs:')
        if type(m.vEIn) is np.ndarray:
            print(m.vEIn)
        else:
            for xn in self.EM.s['Nodz']:
                for xv in self.EM.s['Vec']:
                    aux = m.vEIn[xn, xv].value
                    print("%8.4f " % aux, end='')
                print('')

    def ReadTimeS(self, FileName):
        ''' Read time series for demand and RES '''
        MODEL_JSON = os.path.join(os.path.dirname(__file__), '..', 'json',
                                  FileName)
        Profile_Data = json.load(open(MODEL_JSON))

        # Get number of profiles
        RESTypes = Profile_Data['metadata']['title']
        NoRESP = len(RESTypes)

        # Search for demand profile
        if "Demand" in RESTypes:
            DemandProfiles = Profile_Data['Demand']['Values']
            BusDem = Profile_Data['Demand']['Bus']
            LinkDem = Profile_Data['Demand']['Links']
            NoDemPeriod = len(Profile_Data['Demand']['Period'])
            RESTypes.remove('Demand')
            NoRESP -= 1
        else:
            DemandProfiles = [1]
            NoDemPeriod = 1
            BusDem = 'All'
            LinkDem = 0

        # Check if there is generation data
        if NoRESP == 0:
            # Remove demand data
            LLRESType = 0
            LLRESPeriod = 0
            RESProfs = 0
        else:
            # Location of each type of RES generation technology
            LLRESType = np.zeros((NoRESP, 2), dtype=int)
            NoRES = -1
            xt = -1
            sRes = np.zeros((NoRESP, 3), dtype=int)
            NoLink = 0
            for xr in RESTypes:
                xt += 1
                # Get number of periods and links
                sRes[xt][0] = len(Profile_Data[xr]['Period'])
                sRes[xt][1] = max(Profile_Data[xr]['Links'])
                sRes[xt][2] = len(Profile_Data[xr]['Bus'])
                LLRESType[xt][:] = [NoRES+1, NoRES+sRes[xt][0]]
                NoRES += sRes[xt][0]
                NoLink += sRes[xt][1]

            # Location of data for each period
            LLRESPeriod = np.zeros((NoRES+1, 2), dtype=int)
            RESBus = np.zeros(NoRES, dtype=int)
            RESLink = np.zeros(NoRES, dtype=int)
            xL = -1
            acu = -1
            for xt in range(NoRESP):
                for xp in range(sRes[xt][0]):
                    xL += 1
                    LLRESPeriod[xL][:] = [acu+1, acu+sRes[xt][1]]
                    acu += sRes[xt][1]

            # RES genertaion profiles
            if sRes[0][0]*sRes[0][1] == 1:
                Nohr = len(Profile_Data[RESTypes[0]]['Values'])
            else:
                Nohr = len(Profile_Data[RESTypes[0]]['Values'][0])
            RESProfs = np.zeros((acu+1, Nohr), dtype=float)
            acu = [0, 0, 0]
            xt = -1
            for xr in RESTypes:
                xt += 1
                aux = Profile_Data[xr]['Values']
                RESProfs[acu[0]:acu[0]+sRes[xt][0]*sRes[xt][1]][:] = aux
                acu[0] += sRes[xt][0]*sRes[xt][1]
                for xL in range(sRes[xt][2]):
                    RESBus[acu[1]] = Profile_Data[xr]['Bus'][xL]
                    RESLink[acu[1]] = Profile_Data[xr]['Links'][xL]+acu[2]
                    acu[1] += 1
                acu[2] += sRes[xt][1]

            return (DemandProfiles, NoDemPeriod, BusDem, LinkDem, NoRES,
                    NoRESP, LLRESType, LLRESPeriod, RESProfs, RESBus, RESLink,
                    NoLink, Nohr)

    def run(self, solver=None, subscripts={}):
        ''' Run integrated pyene model '''
        if solver == "pyomo":
            # Build pyomo model
            m = ConcreteModel()
            m = self.build_Mod(m)

            # Run pyomo model
            (m, results) = self.Run_Mod(m)

            return m
        elif solver == "CLP":
            from pyene.engines.pyene_Models import EnergyandNetwork

            ENMod = EnergyandNetwork(self.EN.EM, self.EN.NM, self.EN)
            ENMod.optimisationENM()

            return ENMod
        else:
            from .pyene_Models import models
            if isinstance(self.mod_cython_cpp, models):
                self.mod_cython_cpp.evaluate(subscripts=subscripts)
    
    def save_outputs(self, sim_no=None):
        self.save_solution.save_results(model=self.mod_cython_cpp, sim_no=sim_no)
    
    def update_parameter(self, information):
        return self.mod_cython_cpp.update_parameter(information=information)
    
    def get_moea_variables(self):
        return self.mod_cython_cpp.get_moea_variables()
    
    def get_moea_objectives(self):
        return self.mod_cython_cpp.get_moea_objectives()
    
    def close_output_files(self):
        self.save_solution.close_output_files()

    def Run_Mod(self, m):
        ''' Run pyomo model '''
        # Finalise model
        m = self._AddPyeneCons(m)

        #                          Objective function                         #
        self.OFaux = self._Calculate_OFaux()

        m.OF = Objective(rule=self.OF_rule, sense=minimize)

        # Optimise
        opt = SolverFactory('glpk')
        # Print
        results = opt.solve(m)

        return (m, results)

    def set_Demand(self, index, value):
        ''' Set a demand profile '''
        if index <= self.NM.scenarios['Number']:
            # Do the profiles match?
            value = self.CheckProfile(value)

            aux1 = (index-1)*self.NM.settings['NoTime']
            aux2 = index*self.NM.settings['NoTime']

            self.NM.scenarios['Demand'][aux1:aux2] = value

    def set_GenCoFlag(self, index, value):
        ''' Adjust maximum output of generators '''
        if isinstance(value, bool):
            if value:
                # Maximum capacity
                self.NM.Gen.set_Max(index-1,
                                    self.NM.Gen.get_Max(index-1) /
                                    self.NM.ENetwork.get_Base())
            else:
                # Switch off
                self.NM.Gen.set_Max(index-1, 0)
        else:
            # Adjust capacity
            # TODO: Costs should be recalculated for higher capacities
            value /= self.NM.ENetwork.get_Base()
            if value > self.NM.Gen.get_Max(index-1):
                import warnings
                warnings.warn('Increasing generation capacity is not'
                              ' supported yet')

            self.NM.Gen.set_Max(index-1, value)

    def set_Hydro(self, index, value):
        ''' Set MWh of hydro that are available for a single site '''
        if self.NM.hydropower['Number'] == 1:
            self.EM.Weight['In'][1] = value
        else:
            self.EM.Weight['In'][1][index-1] = value

    def set_HydroPrice(self, index, value):
        ''' Set kWh of hydro that are available for a single site '''
        raise NotImplementedError('Water prices not yet enabled')
        # Adjust self.NM.p['GenLCst']

    def _set_LineCapacityAux(self, value, index, *argv):
        ''' Auxiliary for selecting line parameters '''
        if 'BR_R' in argv:
            for xb in index:
                self.NM.ENetwork.Branch[xb].set_R(value)
        elif 'BR_X' in argv:
            for xb in index:
                self.NM.ENetwork.Branch[xb].set_X(value)
        elif 'BR_B' in argv:
            for xb in index:
                self.NM.ENetwork.Branch[xb].set_B(value)
        else:
            value /= self.NM.ENetwork.get_Base()
            for xb in index:
                self.NM.ENetwork.Branch[xb].set_Rate(value)

    def set_LineCapacity(self, index, value, *argv):
        ''' Adjust maximum capacity of a line - pass BR_R for R/X/B'''
        self._set_LineCapacityAux(value, index, *argv)

    def set_LineCapacityAll(self, value, *argv):
        ''' Adjust capacity of all lines - pass BR_R for R/X/B'''
        index = range(self.NM.ENetwork.get_NoBra())
        self._set_LineCapacityAux(value, index, *argv)

    def set_PumpPrice(self, index, value):
        ''' Set value for water pumped '''
        raise NotImplementedError('Pump prices not yet enabled')
        # Adjust self.NM.p['GenLCst']

    def set_RES(self, index, value):
        '''
        Set PV/Wind profile  
        
        index - beginning from 1
        value - time series
        
        if the index is in self.Overwrite['RESX'], the time series in 'RES'
        will overwrite value
        '''

        # Overwrite value?
        if index in self.Overwrite['RESX']:
            value = self.Overwrite['RES'][self.Overwrite['RESL']
                                          [self.Overwrite['RESX'].index(index)
                                           ]]

        if index <= self.NM.scenarios['NoRES']:
            value = self.CheckProfile(value)

            aux1 = (index-1)*self.NM.settings['NoTime']
            aux2 = index*self.NM.settings['NoTime']

            self.NM.scenarios['RES'][aux1:aux2] = value
            xi = 0
            for xs in range(aux1, aux2):
                self.NM.scenarios['RES'][xs] = \
                    (value[xi]/self.NM.ENetwork.get_Base())
                xi += 1

    def SingleLP(self, ENM):
        ''' Get mathematical model '''
        # Define pyomo model
        mod = ConcreteModel()

        #                                 Sets                                #
        mod = ENM.addSets(mod)

        #                              Parameters                             #
        mod = ENM.addPar(mod)

        #                           Model Variables                           #
        mod = ENM.addVars(mod)

        #                             Constraints                             #
        mod = ENM.addCon(mod)

        return mod


    # TODO: Generalise this functions
    def set_element_status(self, ID=None, status=True, typ=None):
        ''' This function sets the status of the element "typ" to on or off '''
        if typ == "branch":
            if ID:
                self.NM.ENetwork.Branch[ID].data['BR_STATUS'] = status
            else:
                raise ValueError("ID -->{}<--- is not valid to set the status of the branch".format(ID))
        if typ == "generator":
            if ID:
                self.NM.Gen.Conv[ID].data['GEN'] = status
            else:
                raise ValueError("ID -->{}<--- is not valid to set the status of the generator".format(ID))



""" Test the pyeneE engine. """
from pyene.fixtures import testConfig, json_directory
from pyene.engines.pyene import pyeneClass as pe
import os


# Small network
def test_pyeneN_4Bus():
    print('test_pyeneN_4Bus: case4.json')
    conf = testConfig()
    conf.NM.settings['File'] = os.path.join(json_directory(), 'case4.json')
    EN = pe(conf.EN)
    # Initialise model
    (NM, NModel, results) = EN.NSim(conf)
    NM.print(NModel)
    print(NModel.OF.expr())

    assert 0.0001 >= abs(NModel.OF.expr()-21952.5)


# Medium network
def test_pyeneN_14Bus():
    print('test_pyeneN_14Bus: case14.json')
    conf = testConfig()
    conf.NM.settings['File'] = os.path.join(json_directory(), 'case14.json')
    EN = pe(conf.EN)
    # Initialise model
    (NM, NModel, results) = EN.NSim(conf)
    NM.print(NModel)

    assert 0.0001 >= abs(NModel.OF.expr()-7704.910967)


# Small network - Security and losses
def test_pyeneN_4BusSec():
    print('test_pyeneN_4BusSec: case4.json')
    conf = testConfig()
    conf.NM.settings['File'] = os.path.join(json_directory(), 'case4.json')
    conf.NM.settings['Security'] = [2, 3]
    conf.NM.settings['Losses'] = True
    conf.NM.settings['NoTime'] = 2
    conf.NM.scenarios['Demand'] = [1, 1.1]
    conf.NM.settings['Feasibility'] = True

    EN = pe(conf.EN)

    # Initialise model
    (NM, NModel, results) = EN.NSim(conf)
    NM.print(NModel)
    print('Losses')
    Lss01 = NModel.vNLoss[0, 0].value*NM.ENetwork.data['baseMVA']
    Lss02 = NModel.vNLoss[3, 1].value*NM.ENetwork.data['baseMVA']
    print(Lss01, Lss02)
    print(NModel.OF.expr())

    assert (0.0001 >= abs(NModel.OF.expr()-63215.0917360614) and
            0.0001 >= abs(Lss01 - 0.10718868) and
            0.0001 >= abs(Lss02 - 1.56890072)
            )




# distutils: language = c++
# cython: profile=True

from pyensys.engines.cpp_energy_wrapper cimport models

from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool

import copy
    
#cdef class models_cpp_clp():
#    cdef models_cpp* cpp_mod
#
#    def __cinit__(self):
#        self.cpp_mod = new models_cpp()
#    
#    cpdef load_combined_energy_dc_opf_information_cpp(self, LLNAfter, CTGen):
#        cdef int aux_size = len(LLNAfter)
#        cdef vector[vector[int] ] LLNA_cpp
#        cdef vector[int] aux_vec
#        for aux in range(aux_size):
#            aux_vec = LLNAfter[aux, :]
#            LLNA_cpp.push_back(aux_vec)
#        cdef vector[int] CTG_cpp = CTGen
#        self.cpp_mod.load_combined_energy_dc_opf_information(LLNA_cpp, CTG_cpp)
#    
#
#    cpdef add_bus_cpp(self, demand, number, str type_bus):
#        cdef vector[double] P = demand
#        cdef int n = number
#        cdef string t = type_bus.encode('utf-8')
#        self.cpp_mod.add_bus(P, n, t)
#    
#    cpdef add_branch_cpp(self, reactance, resistance, Pmax, \
#        fr, to, number_branch, type_branch):
#        cdef vector[double] r = resistance
#        cdef vector[double] x = reactance
#        cdef vector[double] Pm = Pmax
#        cdef int f = fr
#        cdef int t = to
#        cdef int n = number_branch
#        cdef string s = type_branch.encode('utf-8')
#        self.cpp_mod.add_branch(r, x, Pm, f, t, n, s)
#    
#    cpdef add_generator_cpp(self, P_max, P_min, bus_number, number, type_gen,\
#        fixed_cost, variable_cost, a_pwl, b_pwl, active):
#        cdef vector[double] Pmax = P_max
#        cdef vector[double] Pmin = P_min
#        cdef int bn = bus_number
#        cdef int n = number
#        cdef string t = type_gen.encode('utf-8')
#        cdef double fc = fixed_cost
#        cdef double vc = variable_cost
#        cdef vector[double] a = a_pwl
#        cdef vector[double] b = b_pwl
#        cdef bool is_active = active
#        self.cpp_mod.add_generator(Pmax, Pmin, bn, n, t, fc, vc, a, b, is_active)
#    
#    cpdef set_integer_data_power_system_cpp(self, str name, value):
#        cdef string n = name.encode('utf-8')
#        cdef int val = value
#        self.cpp_mod.set_integer_data_power_system(n, val)
#    
#    cpdef set_continuous_data_power_system_cpp(self, str name, value):
#        cdef string n = name.encode('utf-8')
#        cdef double val = value
#        self.cpp_mod.set_continuous_data_power_system(n, val)
#    
#
#    cpdef load_energy_tree_information_cpp(self, n_nodes, n_trees, LLEB, LLEA, \
#        intake, output, weight):
#        cdef int number_nodes = n_nodes
#        cdef int number_tree = n_trees
#        cdef int aux_size = len(LLEB)
#        cdef vector[vector[int] ] LLEB_cpp
#        cdef vector[int] aux_vec
#        for aux in range(aux_size):
#            aux_vec = LLEB[aux, :]
#            LLEB_cpp.push_back(aux_vec)
#        aux_size = len(LLEA)
#        cdef vector[vector[int] ] LLEA_cpp
#        for aux in range(aux_size):
#            aux_vec = LLEA[aux, :]
#            LLEA_cpp.push_back(aux_vec)
#        cdef vector[double] energy_intake = intake
#        cdef vector[double] energy_output = output
#        cdef vector[double] weight_nodes = weight
#        self.cpp_mod.load_energy_tree_information(number_nodes, number_tree, \
#            LLEB_cpp, LLEA_cpp, energy_intake, energy_output, weight_nodes)
#
#
#    cpdef run_reduced_dc_opf_cpp(self):
#        self.cpp_mod.run_reduced_dc_opf()
#    
#    cpdef run_iterative_reduced_dc_opf_cpp(self):
#        self.cpp_mod.run_iterative_reduced_dc_opf()
#    
#    cpdef run_iterative_reduced_dc_opf_v2_cpp(self):
#        self.cpp_mod.run_iterative_reduced_dc_opf_v2()
#
#    cpdef run_energy_tree_cpp(self):
#        self.cpp_mod.run_energy_tree()
#    
#    cpdef run_combined_energy_dc_opf_r1_cpp(self):
#        self.cpp_mod.run_combined_energy_dc_opf_r1()
#    
#    cpdef run_iterative_combined_energy_dc_opf_cpp(self):
#        self.cpp_mod.run_iterative_combined_energy_dc_opf()
#    
#    cpdef run_iterative_combined_energy_dc_opf_v2_cpp(self):
#        self.cpp_mod.run_iterative_combined_energy_dc_opf_v2()
#
#
#
#    cpdef get_objective_function_cpp(self):
#        return self.cpp_mod.get_objective_function_nm()
#
#    cpdef get_objective_function_em_cpp(self):
#        return self.cpp_mod.get_objective_function_em()
#
#    cpdef get_objective_function_combined_energy_dc_opf_r1_cpp(self):
#        return self.cpp_mod.get_objective_function_combined_energy_dc_opf_r1()
#
#
#
#    cpdef get_energy_tree_solution_cpp(self):
#        cdef vector[double] PartialStorage
#        cdef vector[double] TotalStorage
#        cdef vector[double] InputsTree
#        cdef vector[double] OutputsTree
#        self.cpp_mod.get_energy_tree_solution(PartialStorage, TotalStorage, 
#            InputsTree, OutputsTree)
#        sol_partial = PartialStorage
#        sol_total = TotalStorage
#        sol_inputs = InputsTree
#        sol_outputs = OutputsTree
#        return sol_partial, sol_total, sol_inputs, sol_outputs
#
#    cpdef get_generation_solution_cpp(self):
#        cdef vector[double] generation
#        cdef vector[double] generation_cost
#        self.cpp_mod.get_generation_solution(generation, generation_cost)
#        sol_gen = generation
#        sol_gen_cost = generation_cost
#        return sol_gen, sol_gen_cost
#
#    cpdef get_branch_solution_cpp(self):
#        cdef vector[double] power_flow
#        self.cpp_mod.get_branch_solution(power_flow)
#        sol_power_flows = power_flow
#        return sol_power_flows
#    
#    cpdef get_node_solution_cpp(self):
#        cdef vector[double] angle
#        cdef vector[double] generation_curtailment
#        cdef vector[double] load_curtailment
#        self.cpp_mod.get_node_solution(angle, generation_curtailment, load_curtailment)
#        sol_angle = angle
#        sol_gen_cur = generation_curtailment
#        sol_load_cur = load_curtailment
#        return sol_angle, sol_load_cur, sol_gen_cur


cdef class models_cpp():
    cdef models* cpp_mod

    def __cinit__(self):
        self.cpp_mod = new models()
    
    cpdef create_parameter(self):
        self.cpp_mod.create_parameter()
    
    cpdef load_value(self, type_value, na, val):
        cdef string name = na
        cdef int i_value
        cdef double d_value
        cdef bool b_value
        cdef string s_value
        if type_value == "integer":
            i_value = val
            self.cpp_mod.load_integer(name, i_value, False)
        elif type_value == "double":
            d_value = val
            self.cpp_mod.load_double(name, d_value, False)
        elif type_value == "bool":
            b_value = val
            self.cpp_mod.load_bool(name, b_value, False)
        elif type_value == "string":
            s_value = val.encode('utf-8')
            self.cpp_mod.load_string(name, s_value, False)
        elif type_value == "v_integer":
            for v in val:
                i_value = v
                self.cpp_mod.load_integer(name, i_value, True)
        elif type_value == "v_double":
            for v in val:
                d_value = v
                self.cpp_mod.load_double(name, d_value, True)
        elif type_value == "v_bool":
            for v in val:
                b_value = v
                self.cpp_mod.load_bool(name, b_value, True)
        elif type_value == "v_string":
            for v in val:
                s_value = v.encode('utf-8')
                self.cpp_mod.load_string(name, s_value, True)
        else:
            print("data type *{}* is not valid".format(type_value))
    
    cpdef set_parameter(self, type_var):
        cdef string s_value = type_var
        self.cpp_mod.set_parameter(s_value)

    cpdef initialise(self):
        self.cpp_mod.initialise()
    
    cpdef evaluate(self):
        self.cpp_mod.evaluate()
    
    cpdef return_outputs(self):
        cdef vector[double] values
        cdef vector[int] starts
        cdef vector[ vector[ vector[ string] ] ] characteristics
        self.cpp_mod.return_outputs(values, starts, characteristics)
        return values, starts, characteristics
    
    cpdef update_parameter(self):
        cdef int code
        code = self.cpp_mod.update_parameter()
        return code

    cpdef get_MOEA_variables(self):
        cdef vector[string] IDs
        cdef vector[string] names
        cdef vector[double] min_bnd
        cdef vector[double] max_bnd
        self.cpp_mod.get_MOEA_variables(IDs, names, min_bnd, max_bnd)
        return IDs, names, min_bnd, max_bnd

    cpdef get_moea_objectives(self):
        cdef vector[string] names
        self.cpp_mod.get_moea_objectives(names)
        return names




""" This module contains the FutureDAMS - pyene test case definitions.

These test cases are used to assess the different functionalities of the
engine.

# TODO complete this description once the cases are written.

"""
from .engines.main import pyeneClass, pyeneConfig
from pyomo.core import ConcreteModel
from pyomo.environ import SolverFactory
from .engines.pyeneO import PrintinScreen as PSc


def get_pyene(conf=None):
    """ Get pyene object."""

    return pyeneClass(conf)


def get_pyeneConfig():
    """ Get pyene object."""

    return pyeneConfig()


def test_pyeneAC(config):
    """ Run pypsa based AC power flow """
    config.NM.hydropower['Number'] = 0  # Number of hydropower plants
    
    # Create object
    EN = pyeneClass(config.EN)

    # Initialise with selected configuration
    EN.initialise(config)

    # Fake weather engine
    FileName = 'TimeSeries.json'
    (DemandProfiles, NoDemPeriod, BusDem, LinkDem, NoRES, NoRESP,
     LLRESType, LLRESPeriod, RESProfs, RESBus, RESLink, NoLink,
     Nohr) = EN.ReadTimeS(FileName)
    
    # Profiles
    demandNode = _node()
    demandNode.value = DemandProfiles[0][:]
    demandNode.index = 1
    EN.set_Demand(demandNode.index, demandNode.value)

    # Second scenario
    demandNode = _node()
    demandNode.value = DemandProfiles[1][:]
    demandNode.index = 2
    EN.set_Demand(demandNode.index, demandNode.value)
    
    EN_Interface = EN.getClassInterfaces()

    Model_Pypsa = [EN_Interface.pyene2pypsa(EN.NM, x)[0]
                   for x in range(config.NM.scenarios['NoDem'])]

    for xscen in range(config.NM.scenarios['NoDem']):
        Model_Pypsa[xscen].pf()

    EN.NM.print(Model_Pypsa, range(config.NM.scenarios['NoDem']), None, True)

def test_pyeneE(config):
    """ Execute pyene to access pyeneE - Full json based simulation."""
    EN = pyeneClass(config.EN)
    (EM, EModel, results) = EN.ESim(config)
    EM.print(EModel)


# Network simulation test
def test_pyeneN(config):
    """ Execute pyene to access pyeneN - Full json based simulation."""
    # Create object
    EN = pyeneClass(config.EN)
    # Run model
    (NM, NModel, results) = EN.NSim(config)
    print('\n\nOF: ', NModel.OF.expr())
    NM.print(NModel)
    


# Interaction node
class _node():
    def __init__(self):
        self.value = None
        self.index = None
        self.bus = None
        self.marginal = None
        self.flag = False


# pyene simulation test
def test_pyene(conf):
    """ Execute pyene to access pyeneN - Full json based simulation."""
    # Disable pyeneH
    conf.HM.settings['Flag'] = False

    # Create object
    EN = pyeneClass(conf.EN)

    # Initialise with selected configuration
    EN.initialise(conf)

    # Fake weather engine
    FileName = 'TimeSeries.json'
    (DemandProfiles, NoDemPeriod, BusDem, LinkDem, NoRES, NoRESP,
     LLRESType, LLRESPeriod, RESProfs, RESBus, RESLink, NoLink,
     Nohr) = EN.ReadTimeS(FileName)

    # Single demand node (first scenario)
    demandNode = _node()
    demandNode.value = DemandProfiles[0][:]
    demandNode.index = 1
    EN.set_Demand(demandNode.index, demandNode.value)

    # Second scenario
    demandNode = _node()
    demandNode.value = DemandProfiles[1][:]
    demandNode.index = 2
    EN.set_Demand(demandNode.index, demandNode.value)

    # Several RES nodes
    resInNode = _node()
    for xr in range(EN.NM.RES['Number']):
        resInNode.value = RESProfs[xr][:]
        resInNode.index = xr+1
        EN.set_RES(resInNode.index, resInNode.value)

    # Several hydro nodes
    hydroInNode = _node()
    for xh in range(EN.NM.hydropower['Number']):
        hydroInNode.value = 0
        hydroInNode.index = xh+1
        EN.set_Hydro(hydroInNode.index, hydroInNode.value)

    if conf.EN.solverselection['pyomo']:
        # Run integrated pyene
        mod = ConcreteModel()
        mod = EN.run(mod)

        # Print results
        print('\n\nOF: ', mod.OF.expr())
        # EN.NM.offPrint()
        # EN.NM.Print['Generation'] = True
        # EN.NM.Print['Losses'] = True
        # EN.NM.Print['Flows'] = True
        # EN.NM.Print['GenBus'] = True
        EN.Print_ENSim(mod)

        # Collect unused hydro:
        print()
        hydroOutNode = _node()
        for xh in range(EN.EM.size['Vectors']):
            hydroOutNode.index = xh+1
            print('Hydro %d', hydroOutNode.index, ' left: ',
                  EN.get_Hydro(mod, hydroOutNode.index), ' (',
                  EN.get_HydroMarginal(mod, hydroOutNode.index), ')',
                  EN.get_HydroFlag(mod, hydroOutNode.index))

        # Collect output of pumps
        print()
        pumpNode = _node()
        for xp in range(EN.NM.pumps['Number']):
            pumpNode.index = xp+1
            pumpNode.value = EN.get_Pump(mod, xp+1)
            print('Pump %d: %f' % (pumpNode.index, pumpNode.value))

        # Collect RES spill
        print()
        resOutNode = _node()
        for xp in range(EN.NM.RES['Number']):
            resOutNode.index = xp+1
            resOutNode.value = EN.get_RES(mod, resOutNode)
            print('RES %d: %f' % (resOutNode.index, resOutNode.value))

        # # Collect curtailment per node
        # print()
        # curNode = _node()
        # for xn in range(EN.NM.ENetwork.get_NoBus()):
        #     curNode.bus = xn+1
        #     curNode.value = EN.get_DemandCurtailment(mod, curNode.bus)
        #     print('Dem %d: %f' % (curNode.bus, curNode.value))

        # Collect all curtailment
        print()
        curAll = _node()
        curAll.value = EN.get_AllDemandCurtailment(mod)
        print('Total curtailment:', curAll.value)

    if conf.EN.solverselection['glpk']:
        # Energy model in glpk
        from .engines.pyene_Models import EnergyandNetwork as ENMod
        Model = ENMod(EN.EM, EN.NM, EN)
        Model.optimisationENM()
        # Print results
        print('\n\nOF: ', Model.GetObjectiveFunctionENM())
        Prints = PSc(EN)
        Prints.PrintallResults(Model)

        # Collect unused hydro:
        print()
        OutputsTree = Model.GetOutputsTree()
        EnergybalanceDual = Model.GetEnergybalanceDual()
        aux1 = Model.PenaltyCurtailment/Model.BaseUnitPower
        hydroOutNode = _node()
        for xv in range(Model.NumberTrees):
            hydroOutNode.index = xv+1
            if -EnergybalanceDual[xv, 1] > aux1:
                aux2 = True
            else:
                aux2 = False
            print('Hydro %d', hydroOutNode.index, ' left: ',
                  OutputsTree[xv, 1], ' (',
                  int(EnergybalanceDual[xv, 1]), ')',
                  aux2)

        # Collect output of pumps
        print()
        if Model.NumberPumps > 0:
            pumpNode = _node()
            PumpOperation = Model.GetPumpOperation()
            for xp in range(Model.NumberPumps):
                value = 0
                for xh in Model.LongTemporalConnections:
                    for xt in range(Model.ShortTemporalConnections):
                        value += PumpOperation[xh, xt, xp]
                pumpNode.index = xp+1
                pumpNode.value = value
                print('Pump %d: %f' % (pumpNode.index, pumpNode.value))

        # Collect RES spill
        print()
        if Model.NumberRESGen > 0:
            resOutNode = _node()
            RESGeneration = Model.GetRESGeneration()
            for xp in range(Model.NumberRESGen):
                value = 0
                for xh in Model.LongTemporalConnections:
                    acu = 0
                    for xt in range(Model.ShortTemporalConnections):
                        value += (Model.MaxRESGen[xp]*\
                            Model.RESScenarios[xh, xt, xp] \
                            * Model.BaseUnitPower -
                            RESGeneration[xh, xt, xp])
                resOutNode.index = xp+1
                resOutNode.value = value
                print('RES %d: %f' % (resOutNode.index, resOutNode.value))

        # # Collect curtailment per node
        # print()
        # curNode = _node()
        # for xn in range(EN.NM.ENetwork.get_NoBus()):
        #     curNode.bus = xn+1
        #     curNode.value = EN.get_DemandCurtailment(mod, curNode.bus)
        #     print('Dem %d: %f' % (curNode.bus, curNode.value))

        # Collect all curtailment
        print()
        curAll = _node()
        values = [0, 0]
        value = 0
        if Model.FlagProblem and Model.FlagFeasibility:
            LoadCurtailment = Model.GetLoadCurtailmentNodes()
        else:
            LoadCurtailment = Model.GetLoadCurtailmentSystemED()
        for xs in Model.LongTemporalConnections:
            for xt in range(Model.ShortTemporalConnections):
                if Model.FlagProblem and LoadCurtailment is not None:
                    for k in range(Model.NumberContingencies + 1):
                        for ii in range(Model.NumberNodesPS):
                            value += LoadCurtailment[xs, xt, k, ii]
                            values[EN.NM.ENetwork.Bus[ii].get_LT()] += \
                                LoadCurtailment[xs, xt, k, ii]
                if not Model.FlagProblem and LoadCurtailment is not None:
                    value += LoadCurtailment[xs, xt]
        curAll.value = value, values
        print('Total curtailment:', curAll.value)

def test_pyeneRES(conf):
    '''
    A case study with 2 PV generators located at two buses for 2 representative winter days
    '''
    # TODO THIS NEEDS TO BE ERASED IN A FUTURE RELEASE
    # Initialise simulation methods
    conf.HM.settings['Flag'] = False
    conf.NM.settings['Flag'] = True
    conf.NM.settings['Losses'] = False
    conf.NM.settings['Feasibility'] = True
    conf.NM.settings['NoGenerators'] = 12
    #conf.NM.settings['File'] = os.path.join(json_directory(), 'caseGhana_Sim40_BSec_ManualV02.json')
    conf.NM.settings['NoTime'] = 24
    conf.NM.scenarios['Weights'] = [1 for _ in range(conf.NM.settings['NoTime'])]
    #conf.NM.scenarios['Number'] = 4  # Four scenarios including winter weekday/weekend and summer weekday/weekend
    conf.NM.scenarios['NoDem'] = 2  # Four demand profiles (repeated once) including winter weekday/weekend and summer weekday/weekend

    # RES generators
    conf.NM.RES['Number'] = 2  # Number of RES generators
    conf.NM.RES['Bus'] = [1, 4]  # Location (bus) of generators
    conf.NM.RES['Max'] = [10, 15]  # Generation capacity
    conf.NM.RES['Cost'] = [0.001, 0.001]  # Costs

    # Hydro generators
    conf.NM.hydropower['Number'] = 1  # Number of hydropower plants
    conf.NM.hydropower['Bus'] = [5]  # Location (bus) of hydro
    conf.NM.hydropower['Max'] = [200]  # Generation capacity
    conf.NM.hydropower['Cost'] = [0.001]  # Costs

    # Create object
    EN = pe(conf.EN)

    # Initialise with selected configuration
    EN.initialise(conf)

    ################### case 1 - Winter scenarios ###########################

    # # Demand profile
    # fileName = os.path.join(json_directory(), 'TimeSeries.json')
    # Eprofiles = json.load(open(fileName))
    # EN.set_Demand(1, Eprofiles['Demand']['Values'][0])  # First scenario demand profile - Winter Weekday
    # if conf.NM.scenarios['NoDem'] == 2:
    #     EN.set_Demand(2, Eprofiles['Demand']['Values'][1])  # Second scenario demand profile - Winter Weekend
    #
    # # RES profile
    # resInNode = _node()
    # for xr in range(EN.NM.RES['Number']):
    #     resInNode.value = Eprofiles['PV']['Values'][xr]  # Winter profile for RES generator 1 & 2
    #     resInNode.index = xr + 1
    #     EN.set_RES(resInNode.index, resInNode.value)
    #
    # # Hydro profile
    # hydroInNode = _node()
    # for xh in range(EN.NM.hydropower['Number']):
    #     hydroInNode.value = 1000  # Winter dry seasons MWh
    #     hydroInNode.index = xh + 1
    #     EN.set_Hydro(hydroInNode.index, hydroInNode.value)
    #
    # # Solve the model and generate results
    # m = ConcreteModel()
    # m = EN.run(m)
    # EN.Print_ENSim(m)
    # print('Total curtailment:', EN.get_AllDemandCurtailment(m))
    # print('Spill ', EN.get_AllRES(m))
    # print('OF   : ', m.OF.expr())

    ################### case 2 - Summer scenarios ###########################

    # Demand profile
    fileName = os.path.join(json_directory(), 'TimeSeries.json')
    Eprofiles = json.load(open(fileName))
    EN.set_Demand(1, Eprofiles['Demand']['Values'][4])  # Third scenario demand profile - Summer Weekday
    if conf.NM.scenarios['NoDem'] == 2:
        EN.set_Demand(2, Eprofiles['Demand']['Values'][5])  # Fourth scenario demand profile - Summer Weekend

    # RES profile
    resInNode = _node()
    for xr in range(EN.NM.RES['Number']):
        resInNode.value = Eprofiles['PV']['Values'][xr+4]  # Summer profile for RES generator 1 & 2
        resInNode.index = xr + 1
        EN.set_RES(resInNode.index, resInNode.value)

    # Hydro profile
    hydroInNode = _node()
    for xh in range(EN.NM.hydropower['Number']):
        hydroInNode.value = 10000  # Summer rainy seasons MWh
        hydroInNode.index = xh + 1
        EN.set_Hydro(hydroInNode.index, hydroInNode.value)

    # Solve the model and generate results
    m = ConcreteModel()
    m = EN.run(m)
    EN.Print_ENSim(m)
    print('Total curtailment:', EN.get_AllDemandCurtailment(m))
    print('Spill ', EN.get_AllRES(m))
    print('OF   : ', m.OF.expr())

def hydro_example_tobeerased(conf):
    """ Execute pyene to run the example of baseload"""
    # TODO THIS NEEDS TO BE ERASED IN A FUTURE RELEASE
    # Disable pyeneH
    conf.HM.settings['Flag'] = False
    conf.NM.settings['NoGenerators'] = 13

def hydro_example_tobeerased(conf):
    """ Execute pyene to run the example of baseload - THIS NEEDS TO BE ERASED \
        IN A FUTURE RELEASE"""
    # Disable pyeneH
    conf.HM.settings['Flag'] = False

    conf.NM.settings['Flag'] = True
    conf.NM.settings['Losses'] = False
    conf.NM.settings['Feasibility'] = True
    conf.NM.settings['NoTime'] = 24  # Single period
    conf.NM.scenarios['Weights'] = [1 for _ in range(24)]  # Add weights 
        # to the time steps

    # Adding hydropower plants
    conf.NM.hydropower['Number'] = 2  # Number of hydropower plants
    conf.NM.hydropower['Bus'] = [1, 5]  # Location (bus) of hydro
    conf.NM.hydropower['Max'] = [200, 200]  # Generation capacity
    conf.NM.hydropower['Cost'] = [0.001, 0.001]  # Costs

    # Create object
    EN = pyeneClass(conf.EN)

    # Initialise with selected configuration
    EN.initialise(conf)

    # Single demand node (first scenario)
    demandNode = _node()
    demandNode.value = [0.3333, 0.2623, 0.2350, 0.2240, 0.2240, 
        0.2514, 0.3825, 0.6284, 0.6503, 0.5574, 0.5301, 0.5137, 
        0.5355, 0.5027, 0.4918, 0.5464, 0.7760, 0.9891, 1.0000, 
        0.9399, 0.8634, 0.8142, 0.6885, 0.4918]
    demandNode.index = 1
    EN.set_Demand(demandNode.index, demandNode.value)

    # Several hydro nodes
    hydroInNode = _node()
    for xh in range(EN.NM.hydropower['Number']):
        hydroInNode.value = 10000
        hydroInNode.index = xh+1
        EN.set_Hydro(hydroInNode.index, hydroInNode.value)
    
    # Run integrated pyene
    m = ConcreteModel()
    m = EN.run(m)
    EN.Print_ENSim(m)
    print('Total curtailment:', EN.get_AllDemandCurtailment(m))
    print('Spill ', EN.get_AllRES(m))
    print('OF   : ', m.OF.expr())


def test_pyenetest(mthd):
    '''Test specific functionalities'''
    from .engines.pyeneT import TestClass

    txt = 'test' + str(mthd)
    method_to_call = getattr(TestClass, txt)
    method_to_call(TestClass)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 22 18:44:55 2018

Pyene Hydrology provides methods for mapping the short-term movement of water
throughout specific sections of the system

@author: Dr Eduardo Alejandro Martínez Ceseña
https://www.researchgate.net/profile/Eduardo_Alejandro_Martinez_Cesena
"""
import numpy as np
import networkx as nx
import math
import json
from pyomo.core import Constraint, Var, NonNegativeReals


class pyeneHConfig:
    ''' Default settings used for this class '''
    def __init__(self):
        # Basic settings
        self.settings = {
                'Feas': True,  # Feasibility constraints
                'File': None,  # TODO create input json file
                'Flag': True,  # Add hydrology constraints
                'In': [[0, 1, 600], [1, 3, 800]],  # Fixed water inputs
                'M': 1000,  # Multiplier to  reduce magnitude of time/storage
                'NoTime': 24,  # Number of time steps
                'Penalty': 10000,  # Penalty for feasibility constraints
                'seconds': 3600  # Time resolution
                }
        # River models
        self.rivers = {
                'DepthMax': [4, 4, 4, 4],  # Maximum depth
                'DepthMin': [1, 1, 1, 1],  # MInimum depth
                'From': [1, 2, 4, 4],  # Node - from
                'Length': [1000, 1000, 1000, 1000],  # length (m)
                'Manning': [0.03, 0.03, 0.03, 0.03],  # Mannings 'n
                'Parts': [],  # Divide river into several buckets
                'Ramp': [],  # Add flow variation constraints
                'Share': [1, 1, 0.4, 0.6],  # Links between water flows
                'Slope': [0.0001, 0.0001, 0.0001, 0.0001],  # Slope (m)
                'To': [2, 3, 5, 6],  # Node -to
                'Width': [200, 200, 200, 200]  # width (m)
                }
        # Connections between scenarios
        self.connections = {
                'LinksF': [[0, 0]],  # Connect scen 0:Begin1:End
                'LinksT': [[0, 1]],  # To scen 0:Begin1:End
                'Number': 1  # Number of scenarios
                }
        # Nodes
        self.nodes = {
                'In': [1, 4],  # Nodes with water inflows
                'InCN': [],  # Controlled inputs (Nodes)
                'InCS': [],  # Controlled inputs (Scenarios)
                'Inseries': [],  # Controlled inputs (data)
                'Out': [3, 5, 6]  # Nodes with water outflows
                }
        # Hydropower
        self.hydropower = {
                'Node': [],  # Location (Bus) in the network
                'Efficiency': [],  # Efficiency (pu)
                'Head': [],  # Head (m)
                'Storage': [],  # Storage (MWh)
                'Predefined': [], # Predefined values
                }
        # Pumps
        self.pumps = {
                'From': [],  # Location in water network
                'To': [],  # Location in (or out of) the water network
                'Efficiency': [],  # pu
                'Head': []  # (m)
                }
        # Print flags
        self.Print = {
                'WIn': True,
                'WOut': True,
                'Fup': True,
                'Fdown': True,
                'SoC': True,
                'Feas': True,
                'Stor': True,
                }


class HydrologyClass:
    def __init__(self, obj=None):
        ''' Initialise network class '''
        # Get default values
        if obj is None:
            obj = pyeneHConfig()

        # Copy attributes
        for pars in obj.__dict__.keys():
            setattr(self, pars, getattr(obj, pars))

        # Read input file
        if self.settings['File'] is not None:
            self._Read()

        # sets and parameters used for the mathematical model
        self.s = {}
        self.p = {
                'CapStor': None,  # Capacity of the hydro storage
                'ConInNode': None,  # Node inputs
                'ConOutNode': None,  # Node outputs
                'DeltaT': None,  # Time differential
                'Feas': None,  # Feasibility constraint
                'FeasTime': None,  # Feasibility constraint
                'FLinear': None,  # Time dependent flow approximation
                'FXInput': None,  # Fixed input constraints
                'LLInNode': None,  # Connection Node input
                'LLN2B1': None,  # Conection node2Branch
                'LLN2B2': None,  # Conection node2Branch
                'LLOutNode': None,  # Connection Node Output
                'LLpumpFrom': None,  # Connection 1st node of pump
                'LLpumpTo': None,  # Connection 2nd node of pump
                'LLShare1': None,  # From node to several rivers
                'LLShare2': None,  # From node to several rivers
                'LLStor': None,  # Link node-hydro storage units
                'NoStor': None,  # Number of hydro with storage
                'Qmax': None,  # Maximum flow
                'Qmin': None,  # Minimum flow
                'Penalty': None,  # Penalty for feasibility
                'QLinear': None,  # Volume to flow
                'SoCLinks': None,  # Storage connections (scenarios)
                'River': None,  # Starting point of the scenario (Rivers)
                'TimeStor': None  # Time auxiliary for hydro storage
                }
        # Optimisation
        self.opt = {}

    def _BuildHNetwork(self):
        ''' Build network model '''
        self.networkH = nx.Graph()
        # Adding nodes
        aux = min([min(self.rivers['From']), min(self.rivers['To'])])
        for xn in range(self.nodes['Number']):
            self.networkH.add_node(aux+xn)
        # Adding branches
        for xb in range(self.rivers['Number']):
            self.networkH.add_edge(self.rivers['From'][xb],
                                   self.rivers['To'][xb])

        # Map connections between nodes and branches (non-sequential search)
        NoN2B = self.networkH.number_of_edges()*2  # Number of data points
        LLaux = np.zeros(NoN2B, dtype=int)  # connections (non-sequential)
        LLnext = np.zeros(NoN2B, dtype=int)  # Next connection (non-sequential)
        LLN2B1 = np.zeros(NoN2B, dtype=int)  # connections (sequential)
        # Position of first connection and number of connections
        LLN2B2 = np.zeros((self.networkH.number_of_nodes(), 4), dtype=int)

        x0 = 0  # Initial position (LLaux)
        x1 = 0  # Initial position (branches)
        for xb in range(self.rivers['Number']):
            auxNode = [self.rivers['From'][xb]-1,
                       self.rivers['To'][xb]-1]
            auxX = [3, 1]
            for x2 in range(2):
                # Get next position
                xpos = LLN2B2[auxNode[x2]][auxX[x2]]
                # Initialize if the position is available
                if xpos == 0 and LLN2B2[auxNode[x2]][auxX[x2]-1] == 0:
                    LLN2B2[auxNode[x2]][auxX[x2]] = x0
                    LLN2B2[auxNode[x2]][auxX[x2]-1] = 1
                else:  # Search for next available position
                    while LLnext[xpos] != 0:
                        xpos = LLnext[xpos]
                    # Storing data position
                    LLnext[xpos] = x0
                    LLN2B2[auxNode[x2]][auxX[x2]-1] = \
                        LLN2B2[auxNode[x2]][auxX[x2]-1]+1
                # Storing data point
                LLaux[x0] = x1
                x0 += 1
            x1 += 1

        # Remove the 'next' by arranging the data sequentially
        x0 = 0  # Position LLN2B1
        xacu = 0  # Total number of positions addressed so far
        for x2 in [2, 0]:
            for xn in range(self.networkH.number_of_nodes()):
                # Get first branch position for this node
                xpos = LLN2B2[xn][x2+1]
                if LLN2B2[xn][x2] != 0:
                    # Get other positions if available
                    LLN2B2[xn][x2+1] = xacu
                    xacu += LLN2B2[xn][x2]
                    for x3 in range(LLN2B2[xn][x2]):
                        # Store data sequentially
                        LLN2B1[x0] = LLaux[xpos]
                        xpos = LLnext[xpos]
                        x0 = x0+1

        # Linked list for nodes sending water to different rivers
        LLNodWeight = np.zeros(self.nodes['Number'], dtype=int)
        acu1 = 0
        self.opt['NoShare'] = 0
        for xn in range(self.nodes['Number']):
            if LLN2B2[xn][2] > 1:
                LLNodWeight[acu1] = xn
                acu1 += 1
                self.opt['NoShare'] += LLN2B2[xn][2]-1

        if acu1 == 0:
            self.p['LLShare1'] = []
            self.p['LLShare2'] = []
        else:
            self.p['LLShare1'] = np.zeros((self.opt['NoShare'], 2), dtype=int)
            self.p['LLShare2'] = np.zeros(self.opt['NoShare'], dtype=float)
            xr = 0
            for xn in LLNodWeight[0:acu1]:
                for xb in range(LLN2B2[xn][2]-1):
                    self.p['LLShare1'][xr][:] = \
                        [LLN2B1[LLN2B2[xn][3]+xb], LLN2B1[LLN2B2[xn][3]+xb+1]]
                    self.p['LLShare2'][xr] = \
                        self.rivers['Share'][LLN2B1[LLN2B2[xn][3]+xb+1]] / \
                        self.rivers['Share'][LLN2B1[LLN2B2[xn][3]+xb]]
                    xr += 1

        # LLN2B2 Number of connected branches (Before), position of the first
        self.p['LLN2B1'] = LLN2B1
        self.p['LLN2B2'] = LLN2B2

    def _BuildLLScenario(self):
        ''' Build LL to locate beginning of each scenario '''
        # Create additional input/output nodes if needed
        if len(self.pumps['From']) > 0:
            NoHM = max(max(self.rivers['From']), max(self.rivers['To']))
            aux1 = ['Out', 'In']
            aux2 = ['From', 'To']
            for x in range(2):
                auxf = np.zeros(NoHM, dtype=int)  # Flag
                auxv = np.zeros(NoHM, dtype=int)  # Value
                # Adding hydrology output nodes
                cou = 0
                for xn in self.nodes[aux1[x]]:
                    auxf[xn-1] = 1
                    auxv[cou] = xn
                    cou += 1
                # Adding pumps
                for xn in self.pumps[aux2[x]]:
                    if xn != 0:
                        if auxf[xn-1] == 0:
                            auxf[xn-1] = 1
                            auxv[cou] = xn
                            cou += 1
                self.nodes[aux1[x]] = auxv[0:cou]

        self.nodes['InNumber'] = len(self.nodes['In'])
        self.nodes['OutNumber'] = len(self.nodes['Out'])
        self.connections['set'] = range(self.connections['Number'])

        self.connections['Node'] = np.zeros(self.connections['Number'],
                                            dtype=int)
        self.p['ConRiver'] = np.zeros(self.connections['Number'], dtype=int)
        self.p['ConInNode'] = np.zeros(self.connections['Number'], dtype=int)
        self.p['ConOutNode'] = np.zeros(self.connections['Number'], dtype=int)
        for xh in self.connections['set']:
            self.connections['Node'][xh] = self.nodes['Number']*xh
            self.p['ConRiver'][xh] = self.rivers['Number']*xh
            self.p['ConInNode'][xh] = self.nodes['InNumber']*xh
            self.p['ConOutNode'][xh] = self.nodes['OutNumber']*xh

        # Build LL to flag positions of input.output nodes
        self.p['LLInNode'] = np.zeros((self.nodes['Number'], 2), dtype=int)
        self.p['LLOutNode'] = np.zeros((self.nodes['Number'], 2), dtype=int)
        for xn in range(self.nodes['InNumber']):
            self.p['LLInNode'][self.nodes['In'][xn]-1][:] = [1, xn]
        for xn in range(self.nodes['OutNumber']):
            self.p['LLOutNode'][self.nodes['Out'][xn]-1][:] = [1, xn]

        # Add feasibility constraints
        self.p['Feas'] = np.zeros((self.rivers['Number'],
                                   self.connections['Number']), dtype=int)
        self.p['FeasTime'] = np.zeros(self.settings['NoTime'], dtype=int)
        if self.settings['Feas']:
            self.opt['FeasNo'] = self.rivers['Number']
            self.opt['FeasNoTime'] = self.settings['NoTime']
            cou = 0
            for xs in range(self.connections['Number']):
                for xr in range(self.opt['FeasNo']):
                    self.p['Feas'][xr][xs] = cou
                    cou += 1
            for xt in range(self.opt['FeasNoTime']):
                self.p['FeasTime'][xt] = xt
        else:
            self.opt['FeasNo'] = 1
            self.opt['FeasNoTime'] = 1

    def _BuildParts(self):
        ''' Get number of branches (rivers) and their parts '''
        self.rivers['Number'] = len(self.rivers['From'])
        self.nodes['Number'] = max([max(self.rivers['From']),
                                    max(self.rivers['To'])])

        if len(self.rivers['Parts']) == 0:
            # Default value of one
            self.rivers['Parts'] = np.ones(self.rivers['Number'], dtype=int)
        elif len(self.rivers['Parts']) == 1:
            # preselected value
            aux = self.rivers['Parts'][0]
            self.rivers['Parts'] = np.zeros(self.rivers['Number'], dtype=int)
            for xp in range(self.rivers['Number']):
                self.rivers['Parts'][xp] = aux

        # Add ramp constraints to each part
        Noaux = sum(self.rivers['Parts'])
        aux = len(self.rivers['Ramp'])
        self.rivers['Ramp'] = np.ones(Noaux, dtype=int)
        if aux == 1:
            # Fixed value for all ramps
            aux = self.rivers['Ramp']
            self.rivers['Ramp'] = np.ones(Noaux, dtype=int)
            for xp in range(Noaux):
                self.rivers['Ramp'][xp] = aux
        elif aux > 0:
            # Set of values for ech river section
            aux = self.rivers['Ramp']
            xr = 0
            for xp in range(self.rivers['Number']):
                for xv in range(self.rivers['Parts'][xp]):
                    self.rivers['Ramp'][xr] = aux[xp]
                    xr += 1

        # Create new river buckets
        if Noaux > self.rivers['Number']:
            RiverFrom = np.zeros(Noaux, dtype=int)
            RiverTo = np.zeros(Noaux, dtype=int)
            RiverShare = np.ones(Noaux, dtype=float)
            RiverLength = np.zeros(Noaux, dtype=float)
            RiverSlope = np.zeros(Noaux, dtype=float)
            RiverWidth = np.zeros(Noaux, dtype=float)
            RiverDepthMax = np.zeros(Noaux, dtype=float)
            RiverDepthMin = np.zeros(Noaux, dtype=float)
            RiverManning = np.zeros(Noaux, dtype=float)

            xn = self.nodes['Number']
            xb = 0
            for xr in range(self.rivers['Number']):
                auxL = self.rivers['Length'][xr]/self.rivers['Parts'][xr]
                RiverFrom[xb] = self.rivers['From'][xr]
                RiverShare[xb] = self.rivers['Share'][xr]
                RiverLength[xb] = auxL
                RiverSlope[xb] = self.rivers['Slope'][xr]
                RiverWidth[xb] = self.rivers['Width'][xr]
                RiverDepthMax[xb] = self.rivers['DepthMax'][xr]
                RiverDepthMin[xb] = self.rivers['DepthMin'][xr]
                RiverManning[xb] = self.rivers['Manning'][xr]
                if self.rivers['Parts'][xr] > 1:
                    aux = self.rivers['To'][xr]
                    for xp in range(self.rivers['Parts'][xr]-1):
                        xn += 1
                        RiverTo[xb] = xn
                        xb += 1
                        RiverFrom[xb] = xn
                        RiverLength[xb] = auxL
                        RiverSlope[xb] = self.rivers['Slope'][xr]
                        RiverWidth[xb] = self.rivers['Width'][xr]
                        RiverDepthMax[xb] = self.rivers['DepthMax'][xr]
                        RiverDepthMin[xb] = self.rivers['DepthMin'][xr]
                        RiverManning[xb] = self.rivers['Manning'][xr]
                RiverTo[xb] = self.rivers['To'][xr]
                xb += 1

            self.nodes['Number'] = xn
            self.rivers['Number'] = Noaux
            self.rivers['From'] = RiverFrom
            self.rivers['To'] = RiverTo
            self.rivers['Share'] = RiverShare
            self.rivers['Length'] = RiverLength
            self.rivers['Slope'] = RiverSlope
            self.rivers['Width'] = RiverWidth
            self.rivers['DepthMax'] = RiverDepthMax
            self.rivers['DepthMin'] = RiverDepthMin
            self.rivers['Manning'] = RiverManning

    def _Process(self):
        '''Process information for the optimisation'''
        s = self.rivers['Slope']
        t = self.settings['seconds']
        w = self.rivers['Width']
        n = self.rivers['Manning']
        L = self.rivers['Length']
        M = self.settings['M']
        Q = np.zeros((self.rivers['Number'], 2), dtype=float)
        V = np.zeros((self.rivers['Number'], 2), dtype=float)
        S = np.zeros((self.rivers['Number'], 2), dtype=float)
        D = np.zeros((self.rivers['Number'], 2), dtype=float)
        dt = self.settings['seconds']/self.settings['M']

        # Calculate additional parameters
        txt1 = ['DepthMax', 'DepthMin']
        for xm in range(2):  # Minimum and maximum values
            d = self.rivers[txt1[xm]]
            for xc in range(self.rivers['Number']):
                Q[xc][xm] = w[xc]*d[xc]**(5/3)*s[xc]**0.5/n[xc]  # Flow
                V[xc][xm] = (L[xc]*d[xc]*w[xc])/M  # Water volume
                S[xc][xm] = d[xc]**(2/3)*s[xc]**(1/2)/n[xc]  # Water speed
                D[xc][xm] = min([t*V[xc][xm], L[xc]])  # Distance

        # Adding parameters for ramp constraints
        self.p['Ramp'] = np.zeros(self.rivers['Number'], dtype=float)
        for xc in range(self.rivers['Number']):
            self.p['Ramp'] = (Q[xc][0]-Q[xc][1])*self.rivers['Ramp'][xc]

        # Get linear functios for estimating flows
        # Flows as a function of volume
        QLinear = np.zeros((self.rivers['Number'], 2), dtype=float)
        # Output flows as a function of input flows
        FLinear = np.zeros((self.rivers['Number'], 3), dtype=float)
        for xc in range(self.rivers['Number']):
            # Time dependent constraints for downstream flows
            auxd = S[xc][1]*t
            if auxd == 0:  # Infinite number of buckets required
                auxd = min(S[xc][0]*t, L[xc])
                aux1 = (auxd-0.5*auxd**2/t/S[xc][0])/L[xc]
                aux2 = 0
            else:
                B = math.ceil(L[xc]/auxd)  # Number of buckets required
                dL = L[xc]/B  # Bucket length
                aux1 = min(B, math.floor(S[xc][0]*t/dL))
                aux1 = aux1/B - dL*aux1*(aux1+1)/2/S[xc][0]/t/B
                aux2 = min(B, math.floor(S[xc][1]*t/dL))
                aux2 = aux2/B - dL*aux2*(aux2+1)/2/S[xc][1]/t/B

            # Linear approximation
            FLinear[xc][0] = (Q[xc][0]*aux1+(1-aux1)*Q[xc][1]-Q[xc][1]*aux2 -
                              (1-aux2)*Q[xc][1])/(Q[xc][0]-Q[xc][1])
            FLinear[xc][1] = Q[xc][0]*aux1+(1-aux1)*Q[xc][1] - \
                FLinear[xc][0]*Q[xc][0]
            aux3 = (Q[xc][0]*aux1+(1-aux1)*Q[xc][0]-Q[xc][1]*aux2 -
                    (1-aux2)*Q[xc][0])/(Q[xc][0]-Q[xc][1])
            aux4 = Q[xc][0]*aux1+(1-aux1)*Q[xc][0]-aux3*Q[xc][0]

            # Correction for error
            FLinear[xc][2] = (aux3*Q[xc][0]+aux4 -
                              FLinear[xc][0]*Q[xc][0]-FLinear[xc][1]) / \
                (Q[xc][0]-Q[xc][1])

            # Assigning linear approximation
            # Can the flow go from minimum to maximum?
            if S[xc][0]*t >= L[xc]:
                Vaux = 1
            else:  # Several steps are needed
                Vaux = L[xc]/S[xc][0]/t
            Vaux = max((Q[xc][0]*FLinear[xc][0]-FLinear[xc][1] +
                        FLinear[xc][2]*Q[xc][1])*dt,
                       (-Q[xc][1]*FLinear[xc][0]+FLinear[xc][1] +
                        FLinear[xc][2]*Q[xc][0])*dt)*Vaux

            QLinear[xc][0] = (Q[xc][0]-Q[xc][1])/Vaux
            QLinear[xc][1] = Q[xc][1]-QLinear[xc][0]*V[xc][1]

        self.p['Qmax'] = Q[:, 0]
        self.p['Qmin'] = Q[:, 1]
        self.p['QLinear'] = QLinear
        self.p['FLinear'] = FLinear

    def _Read(self):
        ''' Read input data '''
        # Load file
        mhc = json.load(open(self.settings['File']))

        # River models
        self.rivers['DepthMax'] = mhc['rivers']['DepthMax']
        self.rivers['DepthMin'] = mhc['rivers']['DepthMin']
        self.rivers['From'] = mhc['rivers']['From']
        self.rivers['Length'] = mhc['rivers']['Length']
        self.rivers['Manning'] = mhc['rivers']['Manning']
        self.rivers['Ramp'] = mhc['rivers']['Ramp']
        self.rivers['Share'] = mhc['rivers']['Share']
        self.rivers['Slope'] = mhc['rivers']['Slope']
        self.rivers['To'] = mhc['rivers']['To']
        self.rivers['Width'] = mhc['rivers']['Width']

    def addCon(self, m):
        ''' Add pyomo constraints '''
        if self.settings['Flag']:
            # Constraint on maximum flow downstream
            m.cHQmaxDown = Constraint(self.s['Bra'], self.s['Tim'],
                                      self.s['Sce'], rule=self.cHQmaxDown_rule)
            # Constraint on maximum flow upstream
            m.cHQmaxUp = Constraint(self.s['Bra'], self.s['Tim'],
                                    self.s['Sce'], rule=self.cHQmaxUp_rule)
            # Constraint on minimum flow downstream
            m.cHQminDown = Constraint(self.s['Bra'], self.s['Tim'],
                                      self.s['Sce'], rule=self.cHQminDown_rule)
            # Constraint on minimum flow upstream
            m.cHQminUp = Constraint(self.s['Bra'], self.s['Tim'],
                                    self.s['Sce'], rule=self.cHQminUp_rule)
            # Constraint on minimum flow in output nodes
            m.cHQminOut = Constraint(self.s['OutConstraint'], self.s['Tim'],
                                     self.s['Sce'], rule=self.cHQminOut_rule)
            # Nodal balance
            m.cHNodeBalance = Constraint(self.s['Nod'], self.s['Tim'],
                                         self.s['Sce'],
                                         rule=self.cHNodeBalance_rule)
            # Penalty
            m.cHPenalty = Constraint(rule=self.cHPenalty_rule)
            # River balance
            m.cHRiverBalance = Constraint(self.s['Bra'], self.s['Tim'],
                                          self.s['Sce'],
                                          rule=self.cHRiverBalance_rule)
            # Sharing water among connected rivers
            if self.opt['NoShare'] > 0:
                m.cHWeights = Constraint(self.s['Share'], self.s['Tim'],
                                         self.s['Sce'],
                                         rule=self.cHWeights_rule)
            # Linking SoC in different scenarios
            if self.opt['NoSoCLinks'] > 0:
                m.cHSoCLink = Constraint(self.s['NoSoCLinks'], self.s['Bra'],
                                         rule=self.cHSoCLink_rule)
            # Time dependent constraints on downstream flows
            m.cHQdownTime = Constraint(self.s['Bra'], self.s['Tim'],
                                       self.s['Sce'],
                                       rule=self.cHQdownTime_rule)
            # Fixed inputs
            if self.opt['NoFxInput'] > 0:
                m.cHFixedInput = Constraint(self.s['FXIn'],
                                            rule=self.cHFixedInput_rule)
            m.cHStorMax = Constraint(range(self.p['NoStor']), self.s['Tim'],
                                     self.s['Sce'], rule=self.cHStorMax_rule)
            m.cHStorLink = Constraint(self.s['Stor'],
                                      rule=self.cHStorLink_rule)

            # Ramp constraints (Down)
            m.cHRampDown = Constraint(self.s['Bra'], self.s['Tim'],
                                      self.s['Sce'], rule=self.cHRampDown_rule)
            # Ramp constraints (Up)
            m.cHRampUp = Constraint(self.s['Bra'], self.s['Tim'],
                                    self.s['Sce'], rule=self.cHRampUp_rule)

            # Fixed water inputs
            aux = len(self.nodes['InCN'])
            if aux == 1:  # Add as vector
                m.cHFixedInputV = Constraint(self.s['Tim'],
                                             rule=self.cHFixedInputV_rule)
            elif aux > 1:  # Add as matrix
                m.cHFixedInputM = Constraint(self.s['Tim'], range(aux),
                                             rule=self.cHFixedInputM_rule)
        else:
            # No penalty
            m.cHPenalty = Constraint(expr=m.vHpenalty == 0)

        return m

    def addPar(self, m):
        ''' Adding pyomo parameters '''
        if self.settings['Flag']:
            # Calculate delta t
            self.p['DeltaT'] = self.settings['seconds']/self.settings['M']
            # Fixed input constraints
            self.p['FXInput'] = self.settings['In']
            # Penalty for feasibility
            self.p['Penalty'] = self.settings['Penalty']

        return m

    def addSets(self, m):
        ''' Adding pyomo sets '''
        if self.settings['Flag']:
            self.s['Bra'] = range(self.rivers['Number'])
            self.s['FXIn'] = range(self.opt['NoFxInput'])
            self.s['InNod'] = range(self.nodes['OutNumber'])
            self.s['Nod'] = range(self.nodes['Number'])
            self.s['NoSoCLinks'] = range(self.opt['NoSoCLinks'])
            self.s['Tim'] = range(self.settings['NoTime'])
            self.s['TimP'] = range(self.settings['NoTime']+1)
            self.s['Sce'] = range(self.connections['Number'])
            self.s['Share'] = range(self.opt['NoShare'])
            self.s['Stor'] = range(self.connections['Number'] *
                                   self.p['NoStor'])
            self.s['Pump'] = range(self.connections['Number'] *
                                   self.p['NoPump'])

        return m

    def addVars(self, m):
        ''' Adding pyomo varaibles '''
        if self.settings['Flag']:
            auxr = range(self.connections['Number']*self.rivers['Number'])
            auxin = range(self.connections['Number']*self.nodes['InNumber'])
            auxout = range(self.connections['Number']*self.nodes['OutNumber'])

            # Downstream flow
            m.vHdown = Var(auxr, self.s['Tim'], domain=NonNegativeReals)
            # Feasibility constraint
            m.vHFeas = Var(range(self.opt['FeasNo'] *
                                 self.connections['Number']),
                           range(self.opt['FeasNoTime']),
                           domain=NonNegativeReals, initialize=0.0)
            # Water inputs (Node)
            m.vHin = Var(auxin, self.s['Tim'], domain=NonNegativeReals)
            # Water outputs (Node)
            m.vHout = Var(auxout, self.s['Tim'], domain=NonNegativeReals)
            # State of charge of the river
            m.vHSoC = Var(auxr, self.s['TimP'], domain=NonNegativeReals)
            # Upstream flow
            m.vHup = Var(auxr, self.s['Tim'], domain=NonNegativeReals)
            # Storage for hydropower units
            if self.p['NoStor'] == 0:
                m.vHStor = Var(domain=NonNegativeReals)
            else:
                m.vHStor = Var(self.s['Stor'], self.s['Tim'],
                               domain=NonNegativeReals)

        # Penalty
        m.vHpenalty = Var(domain=NonNegativeReals, initialize=0)

        return m

    def cHFixedInput_rule(self, m, xf):
        ''' Fix some water inputs'''
        return m.vHin[self.p['FXInput'][xf][0], self.p['FXInput'][xf][1]] == \
            self.p['FXInput'][xf][2]

    def cHFixedInputV_rule(self, m, xt):
        ''' Fixed inputs from single node (data in a vector)'''
        return m.vHin[self.nodes['InCN'][0], xt] == self.nodes['Inseries'][xt]

    def cHFixedInputM_rule(self, m, xt, x):
        ''' Fixed inputs from multiple nodes (data in a matrix)'''
        return m.vHin[self.p['ConInNode'][self.nodes['InCS'][x]] +
                      self.nodes['InCN'][x], xt] \
            == self.nodes['Inseries'][x][xt]

    def cHQdownTime_rule(self, m, xr, xt, xh):
        '''
        Downstream flows as a function of upstream flows and water volume'''
        return m.vHdown[self.p['ConRiver'][xh]+xr, xt] == \
            self.p['FLinear'][xr][0]*m.vHup[self.p['ConRiver'][xh]+xr, xt] + \
            self.p['FLinear'][xr][1] - \
            m.vHFeas[self.p['Feas'][xr][xh], self.p['FeasTime'][xt]] + \
            (m.vHSoC[self.p['ConRiver'][xh]+xr, xt]*self.p['QLinear'][xr, 0] +
             self.p['QLinear'][xr, 1]-self.p['Qmin'][xr]) * \
            self.p['FLinear'][xr][2]

    def cHSoCLink_rule(self, m, xs, xr):
        ''' Linking SoC in different scenarios '''
        return m.vHSoC[self.p['SoCLinks'][xs, 0]+xr,
                       self.p['SoCLinks'][xs, 1]] == \
            m.vHSoC[self.p['SoCLinks'][xs, 2]+xr, self.p['SoCLinks'][xs, 3]]

    def cHStorMax_rule(self, m, xn, xt, xh):
        ''' Limits for the storage used by hydropower plants '''
        x = xh*self.p['NoStor']+xn
        return m.vHStor[x, xt] <= self.p['CapStor'][xn+1]

    def cHStorLink_rule(self, m, xn):
        ''' Link first and last periods '''
        return m.vHStor[xn, 0] == m.vHStor[xn, self.settings['NoTime']-1]

    def cHQmaxDown_rule(self, m, xr, xt, xh):
        ''' Constraint on maximum flow upstream'''
        return m.vHdown[self.p['ConRiver'][xh]+xr, xt] <= \
            self.p['Qmax'][xr]+m.vHFeas[self.p['Feas'][xr][xh],
                                        self.p['FeasTime'][xt]]

    def cHQmaxUp_rule(self, m, xr, xt, xh):
        ''' Constraint on maximum flow upstream'''
        return m.vHup[self.p['ConRiver'][xh]+xr, xt] <= self.p['Qmax'][xr] + \
            m.vHFeas[self.p['Feas'][xr][xh], self.p['FeasTime'][xt]]

    def cHQminDown_rule(self, m, xr, xt, xh):
        ''' Constraint on minimum flow upstream '''
        return m.vHdown[self.p['ConRiver'][xh]+xr, xt] >= \
            self.p['Qmin'][xr]-m.vHFeas[self.p['Feas'][xr][xh],
                                        self.p['FeasTime'][xt]]

    def cHQminOut_rule(self, m, xs, xt, xh):
        ''' Constraint on minimum flow for given output nodes '''
        xn = self.p['OutConstraint'][0][xs]
        xr = self.p['OutConstraint'][1][xs]
        return m.vHout[self.p['ConOutNode'][xh]+xn, xt] >= \
            self.p['Qmin'][xr]-m.vHFeas[self.p['Feas'][xr][xh],
                                        self.p['FeasTime'][xt]]

    def cHQminUp_rule(self, m, xr, xt, xh):
        ''' Constraint on minimum flow upstream '''
        return m.vHup[self.p['ConRiver'][xh]+xr, xt] >= self.p['Qmin'][xr] - \
            m.vHFeas[self.p['Feas'][xr][xh], self.p['FeasTime'][xt]]

    def cHNodeBalance_rule(self, m, xn, xt, xh):
        ''' Nodal balance '''
        aux1 = xh*self.p['NoStor']+self.p['LLStor'][xn][1]
        aux2 = range(self.p['LLStor'][xn][0])

        return sum(m.vHin[self.p['ConInNode'][xh] +
                          self.p['LLInNode'][xn, 1], xt]
                   for xb in range(self.p['LLInNode'][xn, 0])) + \
            sum(m.vHdown[self.p['ConRiver'][xh] +
                         self.p['LLN2B1'][self.p['LLN2B2'][xn, 1]+xd], xt]
                for xd in range(self.p['LLN2B2'][xn, 0])) == \
            sum(m.vHup[self.p['ConRiver'][xh] +
                       self.p['LLN2B1'][self.p['LLN2B2'][xn, 3]+xd], xt]
                for xd in range(self.p['LLN2B2'][xn, 2])) +\
            sum(m.vHout[self.p['ConOutNode'][xh] +
                        self.p['LLOutNode'][xn, 1], xt]
                for xb in range(self.p['LLOutNode'][xn, 0])) + \
            sum(m.vHStor[aux1+xs, self.p['TimeStor'][xt][0]]for xs in aux2) - \
            sum(m.vHStor[aux1+xs, self.p['TimeStor'][xt][1]]for xs in aux2)

    def cHPenalty_rule(self, m):
        ''' No penalty '''
        return m.vHpenalty == \
            sum(sum(sum(m.vHFeas[self.p['Feas'][xr][xh],
                                 self.p['FeasTime'][xt]]
                        for xr in self.s['Bra'])*self.p['Penalty']
                    for xt in self.s['Tim'])
                for xh in self.s['Sce'])*1000

    def cHRampDown_rule(self, m, xr, xt, xh):
        ''' Constraint on maximum flow variaiton _UP'''
        return m.vHup[self.p['ConRiver'][xh]+xr, xt] - \
            m.vHup[self.p['ConRiver'][xh]+xr, self.p['FeasTime'][xt]] >= \
            -self.p['Ramp']

    def cHRampUp_rule(self, m, xr, xt, xh):
        ''' Constraint on maximum flow variaiton _UP'''
        return m.vHup[self.p['ConRiver'][xh]+xr, xt] - \
            m.vHup[self.p['ConRiver'][xh]+xr, self.p['FeasTime'][xt]] <= \
            self.p['Ramp']

    def cHRiverBalance_rule(self, m, xr, xt, xh):
        ''' River balance '''
        aux = self.p['ConRiver'][xh]+xr
        return m.vHdown[aux, xt]*self.p['DeltaT'] == m.vHup[aux, xt] * \
            self.p['DeltaT']+m.vHSoC[aux, xt]-m.vHSoC[aux, xt+1]

    def cHWeights_rule(self, m, xw, xt, xh):
        ''' Sharing water from a node among several rivers '''
        return m.vHup[self.p['ConRiver'][xh]+self.p['LLShare1'][xw, 0], xt] * \
            self.p['LLShare2'][xw] == \
            m.vHup[self.p['ConRiver'][xh]+self.p['LLShare1'][xw, 1], xt]

    def initialise(self):
        ''' Initialise engine '''
        if self.settings['Flag']:

            # Get number of branches (rivers) and their parts
            self._BuildParts()

            # Build LL to locate beginning of each scenario
            self._BuildLLScenario()

            # Get settings for each branch (part)
            self._Process()

            # Build network model
            self._BuildHNetwork()

            # Connect the water available in the river at different times
            # e.g., volume at the end and beginning of scenario are the same
            if len(self.connections['LinksF']) > 0:
                aux = np.shape(self.connections['LinksF'])
                if len(aux) == 1:
                    NoSoCLinks = 1
                    val1 = np.zeros((1, 2), dtype=int)
                    val2 = np.zeros((1, 2), dtype=int)
                    val1[0][:] = self.connections['LinksF']
                    val2[0][:] = self.connections['LinksT']
                else:
                    NoSoCLinks = aux[0]
                    val1 = self.connections['LinksF']
                    val2 = self.connections['LinksT']

                SoCLinks = np.zeros((NoSoCLinks, 4), dtype=int)
                aux = [0, self.settings['NoTime']]

                for xL in range(NoSoCLinks):
                    SoCLinks[xL][:] = [self.p['ConRiver'][val1[xL][0]],
                                       aux[val1[xL][1]],
                                       self.p['ConRiver'][val2[xL][0]],
                                       aux[val2[xL][1]]]

            self.p['SoCLinks'] = SoCLinks
            self.opt['NoSoCLinks'] = NoSoCLinks
            self.opt['NoFxInput'] = len(self.settings['In'])

            # Check for storage requirements
            self.p['NoStor'] = len(self.hydropower['Storage'])
            self.p['CapStor'] = np.zeros(self.nodes['Number']+1, dtype=int)
            self.p['TimeStor'] = np.zeros((self.settings['NoTime'], 2),
                                          dtype=int)
            self.p['LLStor'] = np.zeros((self.nodes['Number'], 2), dtype=int)
            if self.p['NoStor'] > 0:
                cou = 0
                for xn in range(self.p['NoStor']):
                    if self.hydropower['Storage'][xn] == 0:
                        self.p['NoStor'] -= 1
                    else:
                        self.p['LLStor'][self.hydropower['Node'][xn]-1][:] = \
                            [1, cou]
                        cou += 1
                        self.p['CapStor'][cou] = \
                            self.hydropower['Storage'][xn]*1000/9.81 / \
                            self.hydropower['Efficiency'][xn] / \
                            self.hydropower['Head'][xn]
                self.p['CapStor'] = self.p['CapStor'][0:cou+1]
            if self.p['NoStor'] > 0:
                for xt in range(1, self.settings['NoTime']):
                    self.p['TimeStor'][xt][0] = xt
                    self.p['TimeStor'][xt][1] = xt-1
                self.p['TimeStor'][0][1] = self.settings['NoTime']-1

            # Adding pumps
            self.p['NoPump'] = len(self.pumps['From'])
            self.p['LLpumpFrom'] = np.zeros((self.nodes['Number'], 2),
                                            dtype=int)
            self.p['LLpumpTo'] = np.zeros((self.nodes['Number'], 2), dtype=int)
            if self.p['NoPump'] > 0:
                # Get node connections
                for x in range(self.p['NoPump']):
                    self.p['LLpumpFrom'][self.pumps['From'][x]-1][:] = [1, x]
                    if self.pumps['To'][x] > 0:
                        self.p['LLpumpTo'][self.pumps['From'][x]-1][:] = [1, x]

            # Node sequence
            self.connections['NodeSeqIn'] = np.zeros(self.nodes['Number'],
                                                     dtype=int)
            for x in range(self.nodes['InNumber']):
                xn = self.nodes['In'][x]-1
                self.connections['NodeSeqIn'][xn] = x+1
            self.connections['NodeSeqOut'] = np.zeros(self.nodes['Number'],
                                                      dtype=int)
            for x in range(self.nodes['OutNumber']):
                xn = self.nodes['Out'][x]-1
                self.connections['NodeSeqOut'][xn] = x+1

            # Output node constraints
            self.p['OutConstraint'] = np.zeros((2, self.nodes['OutNumber']),
                                               dtype=int)
            aux = 0
            for xn in range(self.nodes['OutNumber']):
                node = self.nodes['Out'][xn]
                if node not in self.rivers['From']:
                    xr = 0
                    while self.rivers['To'][xr] != node:
                        xr += 1
                    self.p['OutConstraint'][0][aux] = xn
                    self.p['OutConstraint'][1][aux] = xr
                    aux += 1
            self.s['OutConstraint'] = range(aux)
            self.p['OutConstraint'] = self.p['OutConstraint'][:, 0:aux]

    def OF_rule(self, m):
        ''' Objective function '''
        return sum(sum(sum(m.vHout[self.p['ConOutNode'][xh]+xn, xt]
                           for xn in self.s['InNod'])
                       for xt in self.s['Tim'])
                   for xh in self.s['Sce'])+m.vHpenalty

    def print(self, m, sh=None):
        ''' Print results '''
        if sh is None:
            sh = self.s['Sce']

        if self.settings['Flag']:
            # Nodal results
            for xh in sh:
                print('\n% CASE:', xh)

                if self.Print['WIn']:
                    print("\nWater_In_Node%d=[" % xh)
                    for xn in self.s['Nod']:
                        for xt in self.s['Tim']:
                            if self.p['LLInNode'][xn, 0] != 0:
                                aux = m.vHin[self.p['ConInNode'][xh] +
                                             self.p['LLInNode'][xn, 1],
                                             xt].value
                            else:
                                aux = 0
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['WOut']:
                    print("\nWater_Out_Node%d=[" % xh)
                    for xn in self.s['Nod']:
                        for xt in self.s['Tim']:
                            if self.p['LLOutNode'][xn, 0] != 0:
                                aux = m.vHout[self.p['ConOutNode'][xh] +
                                              self.p['LLOutNode'][xn, 1],
                                              xt].value
                            else:
                                aux = 0
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['Fup']:
                    print("\nFlow_Upstream%d=[" % xh)
                    for xr in self.s['Bra']:
                        for xt in self.s['Tim']:
                            aux = m.vHup[self.p['ConRiver'][xh]+xr, xt].value
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['Fdown']:
                    print("\nFlow_Downstream%d=[" % xh)
                    for xr in self.s['Bra']:
                        for xt in self.s['Tim']:
                            aux = m.vHdown[self.p['ConRiver'][xh]+xr, xt].value
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['SoC']:
                    print("\nRiver%d=[" % xh)
                    for xr in self.s['Bra']:
                        for xt in self.s['TimP']:
                            aux = m.vHSoC[self.p['ConRiver'][xh]+xr, xt].value
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['Stor']:
                    print("\nStorage%d=[" % xh)
                    for xn in self.s['Nod']:
                        for xt in self.s['Tim']:
                            if self.p['LLStor'][xn][0] == 0:
                                aux = 0
                            else:
                                aux = m.vHStor[xh*self.p['NoStor'] +
                                               self.p['LLStor'][xn][1],
                                               self.p['TimeStor']
                                               [xt][0]].value
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

                if self.Print['Feas']:
                    print("\nFlow_Feasibility%d=[" % xh)
                    for xr in self.s['Bra']:
                        for xt in self.s['Tim']:
                            aux = m.vHFeas[self.p['Feas'][xr][xh],
                                           self.p['FeasTime'][xt]].value
                            print("%8.4f " % aux, end='')
                        print()
                    print("];")

    def print_settings(self, m):
        ''' Display input data and results in their original format '''
        auxr = range(self.connections['Number']*self.rivers['Number'])
        auxin = range(self.connections['Number']*self.nodes['InNumber'])
        auxout = range(self.connections['Number']*self.nodes['OutNumber'])

        print('vHSoC = [')
        for xr in auxr:
            for xt in self.s['TimP']:
                print('%f ' % m.vHSoC[xr, xt].value, end='')
            print()
        print('];')

        print('vHup = [')
        for xr in auxr:
            for xt in self.s['Tim']:
                print('%f ' % m.vHup[xr, xt].value, end='')
            print()
        print('];')

        print('vHdown = [')
        for xr in auxr:
            for xt in self.s['Tim']:
                print('%f ' % m.vHdown[xr, xt].value, end='')
            print()
        print('];')

        print('vHin = [')
        for xn in auxin:
            for xt in self.s['Tim']:
                print('%f ' % m.vHin[xn, xt].value, end='')
            print()
        print('];')

        print('vHout = [')
        for xn in auxout:
            for xt in self.s['Tim']:
                print('%f ' % m.vHout[xn, xt].value, end='')
            print()
        print('];')

        print('vHStor = [')
        for xn in self.s['Stor']:
            for xt in self.s['Tim']:
                print('%f ' % m.vHStor[xn, xt].value, end='')
            print()
        print('];')

        print('vHFeas = [')
        for xn in range(self.opt['FeasNo']*self.connections['Number']):
            for xt in range(self.opt['FeasNoTime']):
                print('%f ' % m.vHFeas[xn, xt].value, end='')
            print()
        print('];')

        print('pHConRiver = ', self.p['ConRiver'], end=';\n')
        print('pHConInNode = ', self.p['ConInNode'], end=';\n')
        print('pHConOutNode = ', self.p['ConOutNode'], end=';\n')
        print('pHLLN2B1 = ', self.p['LLN2B1'], end=';\n')

        print('pHLLN2B2 = [')
        for xn in range(self.nodes['Number']):
            for x in range(4):
                print('%d ' % self.p['LLN2B2'][xn][x], end='')
            print()
        print('];')

        print('pHLLInNode = [')
        for xn in range(self.nodes['Number']):
            for x in range(2):
                print('%d ' % self.p['LLInNode'][xn][x], end='')
            print()
        print('];')

        print('pHLLOutNode = [')
        for xn in range(self.nodes['Number']):
            for x in range(2):
                print('%d ' % self.p['LLOutNode'][xn][x], end='')
            print()
        print('];')

        print('pHOptQmax = ', self.p['Qmax'], end=';\n')
        print('pHOptQmin = ', self.p['Qmin'], end=';\n')
        print('pHFeas =', self.p['Feas'], end=';\n')
        print('pHFeasTime = ', self.p['FeasTime'], end=';\n')
        print('pHPenalty =', self.settings['Penalty'], end=';\n')
        print('pHLLShare1 =', self.p['LLShare1'], end=';\n')
        print('pHLLShare2 =', self.p['LLShare2'], end=';\n')

        print('pHQLinear =[')
        for xr in range(self.rivers['Number']):
            for x in range(2):
                print('%f ' % self.p['QLinear'][xr][x], end='')
            print()
        print('];')

        print('pHDeltaT =', self.settings['seconds']/self.settings['M'],
              end=';\n')

        print('pHSoCLinks =[')
        for xs in range(self.opt['NoSoCLinks']):
            for x in range(4):
                print('%d ' % self.p['SoCLinks'][xs][x], end='')
            print()
        print('];')

        print('pHFLinear =[')
        for xr in range(self.rivers['Number']):
            for x in range(3):
                print('%f ' % self.p['FLinear'][xr][x], end='')
            print()
        print('];')
